# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "GraphDelta",
]
from typing import Any, Tuple, Callable
from functools import partial as P

from .. import pyzef
from ._core import *
from . import internals
from .zef_functions import func
from ._ops import *
from .op_structs import ZefOp, LazyValue
from .graph_slice import GraphSlice
from .abstract_raes import Entity, Relation, AtomicEntity

from abc import ABC
class ListOrTuple(ABC):
    pass
for t in [list, tuple]:
    ListOrTuple.register(t)




##############################
# * Description
#----------------------------

# Glossary:
#
# - commands: dictionaries in the constructed GraphDelta which match directly to low-level changes.
# - expressions: high-level things which match to user expectations. There are "complex" expressions and "primitive" expressions.
#
# The general flow to take a bunch of complex exprs, turn them into primitive
# exprs (may increase the number of exprs) and then turn these into commands.
#
# GraphDelta constructor is in three stages:
# First stage: verify input and translate to "simpler" or "fewer" base exprs.
# Second stage: convert exprs to commands and order them
# Orders them by iterating on to-do commands, sorts into bucket of can-be-done-now and to-do-later, applies can-be-done-now, iterate
#
# Logic flow of GraphDelta constructor is:
# - handles nested GraphDeltas (including reassigning unique temporary ids)
# - calls verify_internal_id on each expr (this builds a ID dict for the next step)
# - calls verify_input_el on each expr (this validates that all Z[...] are in the ID dict, as well as the exprs have the expected structure)
# - calls verify_relation_source_target (checks abstract Relations have all source/targets)
# - TODO: call verify no TX or root nodes
# - iteratively calls iteration_step (make_iteration_step)
# - calls verify_and_compact_commands - does ordering
#
# iteration_step is a dispatch on expr type to call functions with _cmds_for prefix (e.g. _cmds_for_instantiable)
# - takes in exprs-to-do ("user-expressions"), finished-commands ("commands") and seen-ids ("ids_present")
# - returns new exprs, new commands and new ids
# - eventually exprs list should "run out" and everything is now a command
# - another analogy: exprs is a "work queue", cmds/ids is an "output list", iteration_step takes from work queue, puts results into output and maybe more jobs into the work queue.
#
#
# Shorthand syntax:
#
# e.g. (a, (b,c,d)) = [ET.Entity, (z, RT.Something, 5)] | g | run
#
# If the user uses the shorthand syntax then the logic first passes through
# `encode`. The function encode only does two additional things to the
# GraphDelta constructor - it assigns IDs where necessary to be able to return
# appropriate results, and prevents a few unusual things (e.g. tagging a
# ZefRef). Most of this function is dedicated to handling relations to be able
# to unpack them into the right structures. Other nodes are left "as is" except
# for obtaining iids to unpack them.
#
#
#
# realise_single_node:
# Used by encode for every top level item and used by GraphDelta constructor and
# encode for pieces of relations. This function exists to both assign an ID and
# convert a complex expression to a primitive one. Only works on "single
# objects", e.g. entity, value, ZefRef, and not relations or tags.
#
# e.g. realise_single_node(ET.Entity) = ("tmp_id_1", ET.Entity["tmp_id_1"])
#
# 
# Performing GraphDelta transaction:
# - works through list of commands, applying then in a low-level transaction onto the graph
# - commands must be properly ordered (done in constructor) before this function is called
# - each command fills in ID dictionary to be returned as receipt.
# - ID dictionary is called "d_raes"
#

# types of commands:
# - instantiate
# - assign_value
# - merge
# - terminate
# - tag
#
# Commands should use keywords of "internal_id" or "origin_rae" to fill in ID dictionary.



########################################################
# * Entrypoint from user code
#------------------------------------------------------

def dispatch_ror_graph(g, x):
    from . import Effect, FX
    if isinstance(x, LazyValue):
        x = collect(x)
        # Note that this could produce a new LazyValue if the input was an
        # assign_value. This is fine.

    if isinstance(x, GraphDelta):
        return Effect({
                "type": FX.TX.Transact,
                "target_graph": g,
                "graph_delta": x
            })
    elif any(isinstance(x, T) for T in {list, tuple, EntityType, AtomicEntityType, ZefRef, EZefRef, ZefOp, QuantityFloat, QuantityInt, LazyValue, Entity, AtomicEntity, Relation}):
        unpacking_template, graph_delta = encode(x)
        # insert "internal_id" with uid here: the unpacking must get to the RAEs from the receipt
        def insert_id_maybe(cmd: dict):
            if 'origin_rae' in cmd:
                if is_a(cmd['origin_rae'], Delegate):
                    internal_id = cmd['origin_rae']
                else:
                    internal_id = uid(cmd['origin_rae'])
                return {**cmd, 'internal_id': internal_id}
            return cmd

        graph_delta_with_ids = GraphDelta([])
        graph_delta_with_ids.commands = [insert_id_maybe(c) for  c in graph_delta.commands]
        return Effect({
                "type": FX.TX.Transact,
                "target_graph": g,
                "graph_delta": graph_delta_with_ids,
                "unpacking_template": unpacking_template,
            })
    raise NotImplementedError(f"'x | g' for x of type {type(x)}")


from ..pyzef import main
main.Graph.__ror__ = dispatch_ror_graph




##############################
# * GraphDelta construction
#----------------------------

class GraphDelta:
    def __repr__(self):
        return ("GraphDelta([\n    "
            +',\n    '.join(self.commands | map[lambda d_raes: str(d_raes)])
            +"\n])")
        
    def __init__(self, elements: tuple):
        generate_id = make_generate_id()    # keeps an internal counter        

        # First extract any nested GraphDeltas commands out
        # Note: using native python filtering as the evaluation engine is too slow

        # nested_cmds = elements | filter[is_a[GraphDelta]] | map[lambda x: x.commands] | concat | collect
        nested_cmds = [x for x in elements if type(x) == GraphDelta] | map[lambda x: x.commands] | concat | collect
        # elements = elements | filter[Not[is_a[GraphDelta]]] | collect
        elements = [x for x in elements if not type(x) == GraphDelta]

        # The nested cmds need to be readjusted for their auto-generated ids 
        # I want to redo this as pure functional style but will do it with side effects for now
        mapping_dict = {}
        def update_ids(cmd, mapping_dict):
            new_cmd = dict(**cmd)
            if "internal_id" in cmd:
                if is_generated_id(cmd["internal_id"]):
                    if cmd["internal_id"] not in mapping_dict:
                        mapping_dict[cmd["internal_id"]] = generate_id()
                    new_cmd["internal_id"] = mapping_dict[cmd["internal_id"]]
            for key in ["source", "target"]:
                if key in cmd:
                    if cmd[key] in mapping_dict:
                        new_cmd[key] = mapping_dict[cmd[key]]
            return new_cmd
                    
        nested_cmds = nested_cmds | map[P(update_ids, mapping_dict=mapping_dict)] | collect
        
        # Obtain all user IDs
        id_definitions = elements | map[obtain_ids] | reduce[merge_no_overwrite][{}] | collect
        # Check that nothing needs a user ID that wasn't provided
        elements | for_each[verify_input_el[id_definitions]]
        elements | for_each[verify_relation_source_target[id_definitions]]

        state_initial = {
            'user_expressions': tuple(elements),
            'commands': nested_cmds,
        }

        def make_debug_output():
            num_done = 0
            next_print = now()+5*seconds
            def debug_output(x):
                nonlocal num_done, next_print
                num_done += 1
                if now() > next_print:
                    print(f"Up to iteration {num_done} - {len(x['user_expressions'])}, {len(x['commands'])}", now())
                    next_print = now() + 5*seconds
            return debug_output
            
        # iterate until all user expressions and generated expressions have become commands
        state_final = (state_initial
                       | iterate[iteration_step[generate_id]]
                       # | map[tap[make_debug_output()]]
                       | take_until[lambda s: len(s['user_expressions'])==0]
                       | last
                       | collect
                       )
        self.commands = verify_and_compact_commands(state_final['commands'])



##############################
# * Validation
#----------------------------

def ensure_not_previously_defined(internal_id: str):
    if internal_id in id_definitions:
        raise KeyError(f"The internal id '{internal_id}' was already defined. Multiple definitions of an internal id are not allowed.")


@func
def obtain_ids(x) -> dict:
    # Return all user IDs for this part of the GraphDelta input.

    # The output of this func (and anytime it is used recursively) must go
    # through merge_no_overwrite to avoid ID clashes.

    ids = {}
    if isinstance(x, LazyValue):
        ids = obtain_ids(x.initial_val)
    elif isinstance(x, ZefOp):
        if is_a(x, instantiated):
            # is there an internal id curried in?
            if len(x.el_ops[0][1]) > 1:
                internal_id = get_curried_arg(x, 1)
                rae_type = get_curried_arg(x, 0)
                ids = {internal_id: rae_type}

        elif is_a(x, merged):
            # is there an internal id curried in?
            if len(x.el_ops[0][1]) > 1:
                internal_id = get_curried_arg(x, 1)
                obj = get_curried_arg(x, 0)
                ids = {internal_id: obj}

        elif is_a(x, assign_value):
            ids = obtain_ids(get_curried_arg(x, 0))

    elif is_a(x, Delegate):
        ids = {x: x}

    # this must be a triple, defining a relation
    elif isinstance(x, ListOrTuple):
        ids = x | map[obtain_ids] | reduce[merge_no_overwrite][{}] | collect

    elif type(x) in [Entity, AtomicEntity, Relation]:
        ids = {uid(x): x}
        if type(x) == Relation:
            if not is_a(x.d["type"][0], RT):
                ids = merge_no_overwrite(ids, obtain_ids(source(x)))
            if not is_a(x.d["type"][2], RT):
                ids = merge_no_overwrite(ids, obtain_ids(target(x)))

    elif type(x) in [ZefRef, EZefRef]:
        ids = {uid(to_ezefref(x)): x}


    # This is an extra step on top of the previous checks
    if type(x) in [Entity, AtomicEntity, Relation, EntityType,
                    AtomicEntityType, RelationType, ZefRef, EZefRef]:
        # Need the lazy value for the RT possibility
        a_id = get_absorbed_id(LazyValue(x))
        if a_id is not None:
            ids = merge_no_overwrite(ids, {a_id: x})

    return ids


@func    
def verify_input_el(x, id_definitions, allow_rt=False, allow_scalar=False):
    # Check each input item to a GraphDelta to validate its form. Uses
    # id_definitions to assert that references to IDs are to things that are
    # defined in the GraphDelta.

    if isinstance(x, LazyValue):
        # This is checking for an (x <= value) format
        if is_a(x.el_ops, assign_value):
            return
        elif is_a(x.el_ops, terminate):
            return
        raise Exception(f"A LazyValue must have come from (x <= value) or (x | terminate) only. Got {x}")

    # Note: just is_a(x, Z) will also mean ZefRefs will be hit
    elif is_a(x, ZefOp) and is_a(x, Z):
        some_id = get_curried_arg(x, 0)

        if some_id not in id_definitions:
            raise KeyError(f"The id '{some_id}' used in Z is not an internal id defined in the GraphDelta init list")
        return

    elif isinstance(x, ZefRef) or isinstance(x, EZefRef):
        return

    elif isinstance(x, ListOrTuple):                         
        if is_valid_relation_template(x):
            if len(x) == 3:
                verify_input_el(x[0], id_definitions, False, True)
                verify_input_el(x[1], id_definitions, True, False)
                verify_input_el(x[2], id_definitions, False, True)
                return
            else:
                return
        for item in x:
            verify_input_el(item, id_definitions, False, False)
        return
    elif isinstance(x, RelationType):
        if allow_rt:
            return
        else:
            raise ValueError(f"Bare RTs without source or target cannot be initialized. You tried to create a {x}.")                    
    elif type(x) in scalar_types:
        if not allow_scalar:
            raise Exception("Direct values are not allowed at the top level of a GraphDelta, as this is likely to indicate a typo, e.g. (ET.Machine, ET.ShouldBeRTNotET, 'name'). Please use explicit value assignment via (AET.String <= 'name') if you really want this behaviour.")
        return
    elif is_valid_single_node(x):
        return

    elif isinstance(x, ZefOp):
        if length(LazyValue(x) | absorbed) > 1:
            # This could be something like AET.Int["asdf"] <= 5
            if x | peel | last | is_a[assign_value]:
                return
        if length(LazyValue(x) | absorbed) != 1:
            raise Exception(f"ZefOp has more than one op inside of it. {x}")
        if is_a(x, terminate):
            return
        if is_a(x, merged):
            return
        if is_a(x, assign_value):
            return
        if is_a(x, tag):
            return
        raise Exception(f"Not allowing ZefOps except for terminate, assign_value, delegate and tag. Got {x}")

    elif is_a(x, Delegate):
        return

    elif type(x) in [Entity, AtomicEntity, Relation]:
        return

    else:
        raise ValueError(f"Unexpected type passed to init list of GraphDelta: {x} of type {type(x)}")

@func    
def verify_relation_source_target(x, id_definitions):
    if type(x) == Relation:
        assert all(u in id_definitions for u in x.d["uids"]), "A Relation doesn't have its corresponding source or target included in the GraphDelta. This is likely because you have an abstract Relation with another Relation as its source/target. These must be included explicitly into the GraphDelta."




##################################################
# * Expr -> cmd conversion
#------------------------------------------------

@func
def iteration_step(state: dict, gen_id)->dict:
    """[summary]

    Args:
        state (dict): of form 
        {
            'user_expressions': elements,
            'commands': (),
        }

    Returns:
        dict: [description]
    """

    # Note: we used to keep track of ids seen here to not merge the same thing
    # twice. However, the user can potentially give two different internal ids
    # to the same merged object, so we can't follow that logic anymore. Instead,
    # the nearly-identical duplicate merges will be contracted later on in the
    # compact commands function.

    exprs = state['user_expressions']
    if len(exprs) == 0:
        return state
    # expr = exprs[0]

    # WARNING: this mutates but is necessary for speed
    if not isinstance(exprs, list):
        exprs = list(exprs)
    expr = exprs.pop()

    new_exprs,new_cmds = dispatch_cmds_for(expr, gen_id)

    # Here are three ways to accomplish the same thing, although the last
    # (and fastest) mutates.

    # return {
    #         'user_expressions': (*exprs[1:], *new_exprs),
    #         'commands': (*state['commands'], *new_cmds),                 # for now this will be a dict to allow quick lookup: the key is the uid / internal id
    #         'ids_present': {*state['ids_present'], *new_ids}
    #     }

    # return {
    #         'user_expressions': exprs[1:] + list(new_exprs),
    #         'commands': state['commands'] + list(new_cmds),
    #         'ids_present': state['ids_present'] | set(new_ids)
    #     }


    exprs.extend(new_exprs)
    cmds = state["commands"]
    cmds.extend(new_cmds)
    return {
            'user_expressions': exprs,
            'commands': cmds,
    }



def dispatch_cmds_for(expr, gen_id):
    if isinstance(expr, LazyValue):
        expr = collect(expr)

    # If we have a lazy value the second time around, then this must be an actual action to perform.
    if isinstance(expr, LazyValue):
        if is_a(expr.el_ops, assign_value):
            return cmds_for_lv_assign_value(expr, gen_id)
        elif is_a(expr.el_ops, terminate):
            return cmds_for_lv_terminate(expr)
        else:
            raise Exception("LazyValue obtained which is not an assign_value or terminate")

    elif isinstance(expr, ZefOp):
        if False:
            pass
        # elif is_a(expr, terminate):
        #     return cmds_for_terminate(expr)
        # elif is_a(expr, assign_value):
        #     return cmds_for_assign_value(expr)
        elif is_a(expr, instantiated):
            return cmds_for_instantiated(expr)
        elif is_a(expr, merged):
            return cmds_for_merged(expr)
        elif is_a(expr, tag):
            return cmds_for_tag(expr)

        raise RuntimeError(f'We should not have landed here, with expr={expr}')
    if is_a(expr, Delegate):
        return cmds_for_delegate(expr)

    d_dispatch = {
        EntityType: cmds_for_instantiable,
        AtomicEntityType: cmds_for_instantiable,

        ZefRef: cmds_for_mergable,
        EZefRef: cmds_for_mergable,
        Entity: cmds_for_mergable,
        AtomicEntity: cmds_for_mergable,
        Relation: cmds_for_mergable,

        tuple: P(cmds_for_tuple, gen_id=gen_id),
        list: P(cmds_for_tuple, gen_id=gen_id),
    }
    if type(expr) not in d_dispatch:
        raise TypeError(f"transform_to_commands was called for type {type(expr)} value={expr}, but no handler in dispatch dict")            
    func = d_dispatch[type(expr)]

    return func(expr)



##################################################
# * Handlers for expr->cmd
#------------------------------------------------

# def cmds_for_complex_expr(x, merged_ids: set, gen_id):
#     iid,exprs = realise_single_node(x, gen_id)
#     return exprs, [], ()

def cmds_for_instantiable(x):
    a_id = get_absorbed_id(x)
    raet = x | without_absorbed | collect
    cmd = {'cmd': 'instantiate', 'rae_type': raet}
    if a_id is not None:
        cmd['internal_id'] =  a_id
        
    return (), [cmd]

def cmds_for_mergable(x):
    origin = origin_rae(x)

    cmd = {"cmd": "merge",
           "origin_rae": origin,
           "internal_ids": []}

    a_id = get_absorbed_id(x)
    if a_id is not None:
        cmd['internal_ids'] += [a_id]

    if is_a(x, ZefRef) or is_a(x, EZefRef):
        if BT(x) == BT.ENTITY_NODE:
            return (), [cmd]

        elif BT(x) == BT.ATOMIC_ENTITY_NODE:
            assign_val_cmds = [] if isinstance(x, EZefRef) else [{
                'cmd': 'assign_value', 
                'value': x | value | collect,
                'internal_id': uid(origin),
                'explicit': False,
                }]

            return (), [cmd] + assign_val_cmds

        elif BT(x) == BT.RELATION_EDGE:
            return (
                (x | source | to_ezefref | collect, x | target | to_ezefref | collect),
                [cmd],
            )
        else:
            raise NotImplementedError(f"Unknown ZefRef type for merging: {BT(x)}")
    elif is_a(x, Entity) or is_a(x, AtomicEntity):
        return (), [cmd]
    elif is_a(x, Relation):
        maybe_src_trg = []
        if not is_a(x.d["type"][0], RT):
            maybe_src_trg.append(source(x))
        if not is_a(x.d["type"][2], RT):
            maybe_src_trg.append(target(x))

        return (
            tuple(maybe_src_trg),
            [cmd],
        )
    else:
        raise NotImplementedError(f"Unknown type for merged[]: {type(x)}")

def cmds_for_tag(x):
    target = x.el_ops[0][1][0]
    cmd = {'cmd': 'tag',
           'tag_name': x.el_ops[0][1][1],
           'force': x.el_ops[0][1][2] if len(x.el_ops[0][1]) >= 3 else False}
    if is_a(target, ZefOp) and is_a(target, Z):
        cmd['internal_id'] = target.el_ops[0][1][0]
    elif is_a(target, ZefRef) or is_a(target, EZefRef):
        cmd['origin_rae'] = origin_rae(target)
    return (), [cmd]
    

def cmds_for_lv_assign_value(x, gen_id: Callable):
    assert isinstance(x, LazyValue)

    assert len(x | peel | collect) == 2
    target,op = x | peel | collect

    assert is_a(op, assign_value)
    val = LazyValue(op) | absorbed | single | collect
    iid,exprs = realise_single_node(target, gen_id)

    return exprs, [{'cmd': 'assign_value', 'value': val, 'explicit': True, "internal_id": iid}]

def cmds_for_lv_terminate(x):
    assert isinstance(x, LazyValue)

    assert len(x | peel | collect) == 2
    target,op = x | peel | collect

    assert is_a(op, terminate)
    cmd = {
        'cmd': 'terminate', 
        'origin_rae': origin_rae(target)
        }
    a_id = get_absorbed_id(LazyValue(op))
    if a_id is not None:
        cmd['internal_id'] = a_id

    return (), [cmd]

def cmds_for_delegate(x):                            
    return (), [{
        'cmd': 'merge', 
        'origin_rae': x,
        }]



def cmds_for_tuple(x: tuple, gen_id: Callable):
    new_exprs = []
    new_cmds = []

    if not is_valid_relation_template(x):
        raise Exception("Not allowing list/tuple nested in GraphDelta unless it belongs to a relation")

    if len(x) == 3:
        if isinstance(x[0], ListOrTuple) and isinstance(x[2], ListOrTuple):
            # Case 4 of is_valid_relation_template
            new_exprs += [(source, x[1], target) for source in x[0] for target in x[2]]
        elif isinstance(x[0], ListOrTuple):
            # Case 3 of is_valid_relation_template
            iid,exprs = realise_single_node(x[2], gen_id)
            new_exprs += exprs
            new_exprs += [(source, x[1], Z[iid]) for source in x[0]]
        elif isinstance(x[2], list):
            # Case 2 of is_valid_relation_template
            iid,exprs = realise_single_node(x[0], gen_id)
            new_exprs += exprs
            new_exprs += [(Z[iid], x[1], target) for target in x[2]]
        else:
            # Case 1 of is_valid_relation_template
            # it's a plain relation triple
            src_id, src_exprs = realise_single_node(x[0], gen_id)
            trg_id, trg_exprs = realise_single_node(x[2], gen_id)

            new_exprs += src_exprs
            new_exprs += trg_exprs

            # This is the only time the relation is created
            rel = x[1]
            if isinstance(rel, RelationType):
                a_id = get_absorbed_id(rel)
                raet = LazyValue(rel) | without_absorbed | collect
                cmd1 = {'cmd': 'instantiate', 'rae_type': raet}
                if a_id is not None:
                    cmd1['internal_id'] =  a_id
            elif isinstance(rel, ZefOp):
                cmd1 = {'cmd': 'instantiate',
                        'rae_type': rel.el_ops[0][1][0],
                        'internal_id': rel.el_ops[0][1][1]}
            cmd1['source'] = src_id
            cmd1['target'] = trg_id

            new_cmds += [cmd1]

    elif len(x) == 2:
        # Case 5 of is_valid_relation_template
        iid,exprs = realise_single_node(x[0], gen_id)
        new_exprs += exprs
        new_exprs += [(Z[iid], item[0], item[1]) for item in x[1]]
    else:
        raise Exception("Shouldn't get here")

    return new_exprs, new_cmds





#### ** Old versions
def cmds_for_instantiated(x: dict):
    internal_id = x.el_ops[0][1][1]
    raet = x.el_ops[0][1][0]
    cmds = []
    if isinstance(raet, QuantityFloat) or isinstance(raet, QuantityInt):
        # This duplicates the logic in realise_single_node a bit here
        val = raet
        if isinstance(val, QuantityFloat):
            raet = getattr(AET.QuantityFloat, val.unit.enum_value)
        else:
            raet = getattr(AET.QuantityInt, val.unit.enum_value)
        exprs += [Z[internal_id] <= val]#{'cmd': 'assign_value', 'value': val, 'explicit': True, 'internal_id': internal_id}]

    cmds += [{'cmd': 'instantiate', 'rae_type': raet, 'internal_id': internal_id}]
    return (), cmds

def cmds_for_merged(x: dict):
    obj = x.el_ops[0][1][0]

    origin = origin_rae(obj)

    base_cmd = {"cmd": "merge",
                "origin_rae": origin,
                "internal_ids": []}
    if len(x.el_ops[0][1]) == 2:
        base_cmd["internal_ids"] += [x.el_ops[0][1][1]]
    else:
        assert len(x.el_ops[0][1]) == 1

    if is_a(obj, ZefRef) or is_a(obj, EZefRef):
        if BT(obj) == BT.ENTITY_NODE:
            return (), [base_cmd]

        elif BT(obj) == BT.ATOMIC_ENTITY_NODE:
            assign_val_cmds = [] if isinstance(obj, EZefRef) else [{
                'cmd': 'assign_value', 
                'value': obj | value | collect,
                'internal_id': uid(origin),
                'explicit': False,
                }]

            return (), [base_cmd] + assign_val_cmds

        elif BT(obj) == BT.RELATION_EDGE:
            return (
                (obj | source | to_ezefref | collect, obj | target | to_ezefref | collect),
                [base_cmd],
            )
        else:
            raise NotImplementedError(f"Unknown ZefRef type for merging: {BT(obj)}")
    elif is_a(obj, Entity) or is_a(obj, AtomicEntity):
        return (), [base_cmd]
    elif is_a(obj, Relation):
        maybe_src_trg = []
        if not is_a(obj.d["type"][0], RT):
            maybe_src_trg.append(source(obj))
        if not is_a(obj.d["type"][2], RT):
            maybe_src_trg.append(target(obj))

        return (
            tuple(maybe_src_trg),
            [base_cmd],
        )
    else:
        raise NotImplementedError(f"Unknown type for merged[]: {type(obj)}")
        



##############################################################
# ** Complex->primitve expr utils
#------------------------------------------------------------

def is_valid_single_node(x):
    if type(x) in scalar_types:
        return True
    if isinstance(x, ZefRef) or isinstance(x, EZefRef):
        return True
    if is_a(x, Z):
        return True
    if isinstance(x, EntityType):
        return True
    if isinstance(x, RelationType):
        return True
    if isinstance(x, AtomicEntityType):
        return True
    return False

def is_valid_relation_template(x):
    # Check for the relation templates:
    # Either:
    # 1. (a, RT.x, b)
    # 2. (a, RT.x, [b,c,d,e,...])
    # 3. ([a,b,c,...], RT.x, d)
    # 4. ([a,b,c,...], RT.x, [d,e,f,...])
    # 5. (a, [(RT.x, b), (RT.y, c)])
    # Note: tuple/list can be used interchangably
    # Note: 5. deliberately does not have its symmetric counterpart. This could be introduced later.
    if any(isinstance(item, RelationType) for item in x):
        # Cases 1-4 above
        if len(x) != 3:
            raise Exception("A list has an RT but isn't 3 elements long")
        if not isinstance(x[1], RelationType):
            raise Exception("A list has an RT but it isn't in the second position")
        # Note: if there are any lists involved, we cannot have a given ID as it
        # would have to be given to multiple instantiated relations.
        if not is_a(x[1], RT) and (isinstance(x[0], ListOrTuple) or isinstance(x[2], ListOrTuple)):
            raise Exception("An RT with an internal name is not allowed with multiple sources or targets")
        return True
    elif len(x) == 2 and isinstance(x[1], ListOrTuple):
        return all(isinstance(item, ListOrTuple) and len(item) == 2 and isinstance(item[0], RelationType) and is_valid_single_node(item[1]) for item in x[1])
    return False


@func
def realise_single_node(x, gen_id):
    # Take something that should refer to a single node, i.e. a RAE or a scalar
    # to be turned into a RAE, or a reference to a RAE, and return a version
    # with an explicit ID and the ID itself.
    #
    # For example
    #
    # realise_single_node(ET.Machine["a"])
    # will return
    # ("a", ET.Machine["a"])
    #
    # realise_single_node(ET.Machine)
    # might return
    # ("12323425346", ET.Machine["12323425346"])


    #     5 different cases we need to deal with:
    # 4) A ZefRef / EZefRef to an existing RAE is specified. Generate the merge command if it is not present yet
    # 5) Z['my_temp_id4']       The temp id may not exist yet (could come from another edge creation). Make sure it is checked to exist at the end
    # 3) e.g. if a plain value was specified, create an AE and assign the value
    # 1) e.g. if ET.Dog is specified as the source: definitely create it
    # 2) e.g. if ET.Dog['Rufus'] was specified, that will be there as a dict. It is the command to create it and register the temp id

    # First case of removing lazy values
    if isinstance(x, LazyValue):
        x = collect(x)

    # Now this is a check for whether we are an assign_value
    if isinstance(x, LazyValue):
        target,op = x | peel | collect
        if is_a(op, assign_value):
            iid,exprs = realise_single_node(target, gen_id)
            exprs = exprs + [LazyValue(Z[iid]) | op]
        elif is_a(op, terminate):
            iid = origin_uid(target)
            exprs = [x]
        else:
            raise Exception(f"Don't understand LazyValue type: {op}")
    elif isinstance(x, EntityType) or isinstance(x, AtomicEntityType):
        a_id = get_absorbed_id(x)
        if a_id is None:
            iid = gen_id()
            exprs = [x[iid]]
        else:
            iid = a_id
            exprs = [x]
    elif isinstance(x, ZefRef) or isinstance(x, EZefRef):
        if internals.is_delegate(x):
            d = to_delegate(x)
            exprs = [d]
            iid = d
        else:
            exprs = [x]
            iid = origin_uid(x)
    elif type(x) in [Entity, AtomicEntity, Relation]:
        exprs = [merged[x]]
        iid = origin_uid(x)
    elif type(x) in scalar_types:
        iid = gen_id()
        aet = map_scalar_to_aet_type[type(x)](x)
        exprs = [aet[iid], LazyValue(Z[iid]) | assign_value[x]]
    elif isinstance(x, ZefOp):
        params = LazyValue(x) | peel | first | second
        if is_a(x, Z):
            iid = params | first | collect
            # No expr to perform
            exprs = []
        elif is_a(x, instantiated):
            if length(params) == 1:
                # We need to create a new id for this object
                x = x[gen_id()]
                params = LazyValue(x) | peel | first | second
            iid = params | second | collect
            exprs = [x]
        elif is_a(x, merged):
            # Note: we always use the origin_uid here, even if there's an
            # internal id curried in. Simpler to later order operations.
            iid = origin_uid(params | first | collect)
            exprs = [x]
        # elif is_a(x, assign_value):
        #     iid = origin_uid(params | first | collect)
        #     exprs = [x]
        # elif is_a(x, terminate):
        #     iid = origin_uid(params | first | collect)
        #     exprs = [x]
        else:
            raise NotImplementedError(f"Can't pass zefops to GraphDelta: for {x}")
    elif is_a(x, Delegate):
        iid = x
        exprs = [x]
    else:
        raise TypeError(f'in GraphDelta encode step: for {type(x)=}')

    return iid, exprs



##############################
# * Ordering commands
#----------------------------

def verify_and_compact_commands(cmds: tuple):                
    aes_with_explicit_assigns = (cmds
                                 | filter[lambda d: d["cmd"] == "assign_value"]
                                 | filter[lambda d: d["explicit"]]
                                 | map[lambda d: d["internal_id"]]
                                 | collect
                                 )
    def is_unnecessary_automatic_merge_assign(d):
        return (d["cmd"] == "assign_value"
                and not d["explicit"]
                and d["internal_id"] in aes_with_explicit_assigns)

    # make sure if multiple assignment commands exist for the same AE, that their values agree
    cmds = (cmds | group_by[get['cmd']]
            | map[match_apply[
                        (first | equals["assign_value"],
                         second | filter[Not[is_unnecessary_automatic_merge_assign]]
                         | validate_and_compress_unique_assignment),
                        (first | equals["merge"], second | combine_internal_ids_for_merges),
                        (first | equals["terminate"], second | combine_terminates),
                        # Just pack things back in for other cmd types
                        (always[True], second)
            ]]
            | concat
            | collect)

    sorted_cmds = tuple(cmds 
                        | sort[command_ordering_by_type] 
                        | collect
                        )

    def make_debug_output():
        num_done = 0
        next_print = now()+5*seconds
        def debug_output(x):
            nonlocal num_done, next_print
            num_done += 1
            if now() > next_print:
                print(f"Up to compacting {num_done} - {len(x['state']['input'])}, {len(x['state']['output'])}", now())
                next_print = now() + 5*seconds
        return debug_output

    state_initial = {
        'input': sorted_cmds,
        'output': (),
        'known_ids': set(),
    }
    state_final = (
        {"state": state_initial, "num_changed": -1}
        | iterate[resolve_dag_ordering_step] 
        # | tap[make_debug_output()]
        | take_until[lambda s: s["num_changed"] == 0]
        # | tap[lambda x: print("After take_until")]
        | last
        | get["state"]
        | collect
    )
    # print("Done state_final", now())
    if len(state_final['input']) > 0:
        import json
        print(json.dumps(state_final, indent=4, default=repr))
        raise NotImplementedError("Error constructing GraphDelta: instantiation order iteration did not converge. Probably there is a circular dependency in the required imperative instantiation order between inter-dependent relations. This is a valid GraphDelta in principle, but currently not implemented in zefDB.")
    ordered_cmds = state_final['output']
    return ordered_cmds

@func
def validate_and_compress_unique_assignment(cmds):
    @func
    def check_list_is_distinct(cmds):
        values = cmds | map[get["value"]] | distinct | collect
        if length(values) == 1:
            return cmds[0]
        raise ValueError(f'There may be at most one assignment commands for each AE. There were multiple for assignment to {get_id(cmds[0])!r} with values {values}')

    return (cmds
            | group_by[get_id]
            | map[second | check_list_is_distinct])
            

@func
def combine_internal_ids_for_merges(cmds):
    def combine_ids(x, y):
        assert x["origin_rae"] == y["origin_rae"]
        return {**x,
                "internal_ids": x["internal_ids"] + y["internal_ids"]}
    
    return (cmds
              # Convert each item to a dictionary using the origin_rae as key
            | map[lambda x: {x["origin_rae"]: x}]
              | reduce[merge_with[combine_ids]][{}]
              # Now we should have one big dict with each command
            | values)

@func
def combine_terminates(cmds):
    @func
    def check_list_is_distinct(cmds):
        # cmds = cmds | distinct | collect
        # if length(cmds) == 1:
        #     return cmds[0]
        if not cmds | map[equals[first(cmds)]] | all | collect:
            ids = cmds | map[get["internal_id"][None]] | collect
            raise ValueError(f'There may be at most one id for a terminated ZefRef. There were multiple for assignment to {cmds[0]["origin_rae"]!r} with ids {ids}')
        return cmds[0]

    return (cmds
            | group_by[get["origin_rae"]]
            | map[second | check_list_is_distinct])



def command_ordering_by_type(d_raes: dict) -> int:
    """we want some standardized order of the output to simplify value-based
        comparisons and other operations for graph deltas"""
        
    if d_raes['cmd'] == 'merge':
        if isinstance(d_raes['origin_rae'], Relation): return 0.5
        else: return 0
    if d_raes['cmd'] == 'instantiate':
        if isinstance(d_raes['rae_type'], EntityType): return 1
        if isinstance(d_raes['rae_type'], AtomicEntityType): return 2
        if isinstance(d_raes['rae_type'], RelationType): return 3
        return 4                                            # there may be {'cmd': 'instantiate', 'rae_type': AET.Bool}
    if d_raes['cmd'] == 'assign_value': return 5
    if d_raes['cmd'] == 'terminate' and isinstance(d_raes['origin_rae'], Relation): return 6
    if d_raes['cmd'] == 'terminate' and isinstance(d_raes['origin_rae'], Entity): return 7
    if d_raes['cmd'] == 'terminate' and isinstance(d_raes['origin_rae'], AtomicEntity): return 8
    if d_raes['cmd'] == 'terminate' and is_a(d_raes['origin_rae'], Delegate): return 9
    if d_raes['cmd'] == 'tag': return 10
    else: raise NotImplementedError(f"In Sort fct for {d_raes}")

def get_id(cmd):
    if "internal_ids" in cmd:
        raise Exception("Not allowed to ask for id of command that could have multiple ids")
    elif 'internal_id' in cmd:
        return cmd['internal_id']
    elif 'origin_rae' in cmd:
        if is_a(cmd['origin_rae'], Delegate):
            return cmd['origin_rae']
        return uid(cmd['origin_rae'])
    else:
        return None

def get_ids(cmd):
    ids = []
    if cmd['cmd'] == 'merge':
        ids += cmd["internal_ids"]
        if is_a(cmd['origin_rae'], Delegate):
            ids += [cmd['origin_rae']]
        else:
            ids += [uid(cmd['origin_rae'])]
    else:
        this_id = get_id(cmd)
        if this_id is not None:
            ids += [this_id]
    return ids

def resolve_dag_ordering_step(arg: dict)->dict:
    # arg is "state" + "num_changed"
    state = arg["state"]
    ids = state['known_ids']        
    
    def can_be_executed(cmd):        
        if cmd['cmd'] == 'instantiate' and isinstance(cmd['rae_type'], RelationType) and (cmd['source'] not in ids or cmd['target'] not in ids):
            return False
        if cmd['cmd'] == 'merge' and isinstance(cmd['origin_rae'], Relation) and (cmd['origin_rae'].d["uids"][0] not in ids or cmd['origin_rae'].d["uids"][2] not in ids):
            return False
        if cmd['cmd'] == 'terminate' and any(get_id(cmd) == get_id(other) for other in state['input'] if other['cmd'] != 'terminate'):
            return False
        return True        
    (_,can), (_,cannot) = state['input'] | group_by[can_be_executed][[True, False]] | collect
    return {
        "state": {
            'input': tuple(cannot),
            'output': (*state['output'], *can),
            'known_ids': {*state['known_ids'], *(can | map[get_ids] | concat)},
        },
        "num_changed": len(can) > 0
    }
    
        

########################################################
# * Shorthand syntax handlers
#------------------------------------------------------



def encode(xx):
    """
    This function is invoked if one writes "[ET.Dog, (ET.Person, ET.Aardvark)] | g"
    It returns a GraphDelta with ids assigned, as well as a schema
    with the same structure as the nested input arrays, but the ids as placeholders

    Args:
        xx ([type]): [description]

    Returns:
        [type]: Tuple, 
    """
    gd_exprs = []
    
    gen_id = make_generate_id()

    # - encode works recursively
    # - step handles each recursive step
    #
    # logic:
    # - if step gets a valid relation - lots of logic to extract correct structure
    # - else if step gets a list/tuple - recurse on each item
    # - else must be a single item - use realise_single_node
    #
    # - plus some helpful error handling
    #
    # - note: heavy lifting is passed to GraphDelta constructor - this function just adds IDs, nothing else
    

    def step(x, allow_scalar):
        if isinstance(x, ListOrTuple):
            if is_valid_relation_template(x):
                if len(x) == 2:
                    ent = step(x[0], True)
                    triples = [step((Z[ent], *tup), False) for tup in x[1]]
                    doubles = [x[1:] for x in triples]

                    return (ent, doubles)

                if len(x) == 3:
                    if isinstance(x[0], ListOrTuple):
                        assert is_valid_single_node(x[2])
                        item = step(x[2], True)
                        triples = [step((source, x[1], Z[item]), False) for source in x[0]]
                        sources = [x[0] for x in triples]
                        # Note: have to return None for the relation, as there are actually multiple relations
                        return (sources, None, item)

                    if isinstance(x[2], ListOrTuple):
                        assert is_valid_single_node(x[0])
                        item = step(x[0], True)
                        triples = [step((Z[item], x[1], target), False) for target in x[2]]
                        targets = [x[2] for x in triples]
                        # Note: have to return None for the relation, as there are actually multiple relations
                        return (item, None, targets)

                    # it's a relation triple
                    src_id = step(x[0], True)
                    rel_id = gen_id()
                    trg_id = step(x[2], True)
                    gd_exprs.append( (Z[src_id], x[1][rel_id], Z[trg_id]) )

                    return (src_id, rel_id, trg_id)
                raise Exception("Shouldn't get here")
            else:
                # recurse: return isomorphic structure with IDs
                return tuple((step(el, False) for el in x))
        
        # These next few ifs are for checks on syntax only
        if type(x) in scalar_types:
            if not allow_scalar:
                raise Exception("Scalars are not allowed on their own to avoid accidental typos such as (ET.X, ET.Y, 'z') when (ET.X, RT.Y, 'z') is meant. If you want this behaviour, then create an explicit AET, i.e. (AET.String <= 'z').")
        
        if isinstance(x, ZefOp):
            if is_a(x, instantiated):
                raise Exception("Not allowed to pass internal names to RAEs in shorthand syntax")

        iid,exprs = realise_single_node(x, gen_id)
        gd_exprs.extend(exprs)
        return iid

    step_res = step(xx, False)
    return step_res, GraphDelta(gd_exprs)


def unpack_receipt(unpacking_template, receipt: dict):
    def step(x):
        if isinstance(x, tuple):
            return tuple((step(el) for el in x))
        if isinstance(x, list):
            return [step(el) for el in x]
        return receipt[x] if isinstance(x, str) or is_a(x, uid) else x
    return step(unpacking_template)



##################################################
# * Performing transaction
#------------------------------------------------

def perform_transaction_gdelta(g_delta, g: Graph):
    d_raes = {}  # keep track of instantiated RAEs with temp ids            
    try:
        with Transaction(g) as tx_now:
            # TODO: Have to change the behavior of Transaction(g) later I suspect
            frame_now = GraphSlice(tx_now)
            d_raes['tx'] = tx_now
            for i,cmd in enumerate(g_delta.commands):

                zz = None
                
                # print(f"{i}/{len(g_delta.commands)}: {g.graph_data.write_head * 16 / 1024 / 1024} MB")
                if cmd['cmd'] == 'instantiate' and type(cmd['rae_type']) in {EntityType, AtomicEntityType}:
                    zz = instantiate(cmd['rae_type'], g)
                
                elif cmd['cmd'] == 'instantiate' and type(cmd['rae_type']) in {RelationType}:
                    zz = instantiate(to_ezefref(d_raes[cmd['source']]), cmd['rae_type'], to_ezefref(d_raes[cmd['target']]), g) | in_frame[frame_now] | collect
                
                elif cmd['cmd'] == 'assign_value':
                    this_id = cmd['internal_id']
                    zz = d_raes.get(this_id, None)
                    if zz is None:
                        zz = most_recent_rae_on_graph(this_id, g)
                    assert zz is not None
                    if cmd['value'] is not  None:
                        if zz | value | collect != cmd['value']:
                            # print("Assigning value of ", cmd['value'], "to a", AET(z))
                            internals.assign_value_imp(zz, cmd['value'])
                    zz = now(zz)
                    
                elif cmd['cmd'] == 'merge':
                    # It is either an instance (with an 'origin_rae_uid' specified)
                    # or a delegate[...]
                    if is_a(cmd['origin_rae'], Delegate):
                        d = cmd['origin_rae']
                        zz = internals.delegate_to_ezr(d, g, True, 0)
                        d_raes[d] = zz
                    else:
                        candidate = most_recent_rae_on_graph(uid(cmd['origin_rae']), g)
                        if candidate is not None:
                            # this is already on the graph. Just assert and move on
                            assert rae_type(cmd['origin_rae']) == rae_type(candidate)
                            zz = candidate
                        else:
                            origin_rae_uid = uid(cmd['origin_rae'])
                            if isinstance(cmd['origin_rae'], Entity):
                                zz = internals.merge_entity_(
                                    g, 
                                    rae_type(cmd['origin_rae']), 
                                    origin_rae_uid.blob_uid,
                                    origin_rae_uid.graph_uid,
                                )
                            elif isinstance(cmd['origin_rae'], AtomicEntity):
                                zz = internals.merge_atomic_entity_(
                                    g, 
                                    rae_type(cmd['origin_rae']),
                                    origin_rae_uid.blob_uid,
                                    origin_rae_uid.graph_uid,
                                )
                            elif isinstance(cmd['origin_rae'], Relation):
                                src_origin_uid,_,trg_origin_uid = cmd['origin_rae'].d["uids"]
                                z_src = d_raes.get(src_origin_uid, most_recent_rae_on_graph(src_origin_uid, g))                                    
                                z_trg = d_raes.get(trg_origin_uid, most_recent_rae_on_graph(trg_origin_uid, g))                                    

                                assert z_src is not None
                                assert z_trg is not None                                    
                                zz = internals.merge_relation_(
                                    g, 
                                    rae_type(cmd['origin_rae']),
                                    to_ezefref(z_src),
                                    to_ezefref(z_trg),
                                    origin_rae_uid.blob_uid,
                                    origin_rae_uid.graph_uid,
                                )
                            else:
                                raise NotImplementedError

                        zz = now(zz)
                        if 'origin_rae' in cmd:
                            d_raes[uid(cmd['origin_rae'])] = zz
                
                elif cmd['cmd'] == 'terminate':
                    if uid(cmd['origin_rae']) not in g:
                        # raise KeyError(f"terminate {cmd['origin_rae']} called, but this RAE is not know to graph {uid(g)}")
                        # This is no longer an error, it should be ignored silently.
                        pass
                    else:
                        zz = most_recent_rae_on_graph(uid(cmd['origin_rae']), g)
                        if zz is not None:
                            from ..pyzef import zefops as pyzefops
                            pyzefops.terminate(zz)

                            zz = now(zz, allow_tombstone)
                elif cmd['cmd'] == 'tag':
                    if 'origin_rae' in cmd:
                        if uid(cmd['origin_rae']) not in g:
                            raise KeyError(f"tag {cmd['origin_rae']} called, but this RAE is not know to graph {g|uid}")
                        zz = most_recent_rae_on_graph(uid(cmd['origin_rae']), g)
                    else:
                        zz = d_raes[cmd['internal_id']]
                    # Go directly to the tagging instead of through another effect
                    pyzef.main.tag(zz, cmd['tag_name'], cmd['force'])
                else:
                    raise RuntimeError(f'---------Unexpected case in performing graph delta tx: {cmd}')

                for this_id in get_ids(cmd):
                    d_raes[this_id] = zz

    except Exception as exc:        
        raise RuntimeError(f"Error executing graph delta transaction {exc=}") from exc
        
    return d_raes    # the transaction receipt






def perform_transaction(x, g: Graph):
    """
    x can be 
        1) a GraphDelta
        2) a list/tuple of GraphDeltas
        3) s horthand type for txs: single ET/AET/triple/ list thereof
        
    if a tuple of GraphDelta is piped in, perform all of them within the same transaction."""
    if isinstance(x, GraphDelta):
        return perform_transaction_gdelta(x, g)
    elif isinstance(x, list) or isinstance(x, tuple):
        with Transaction(g):
            res = tuple((perform_transaction_gdelta(el, g) for el in x))
        return res
    else:
        raise TypeError(f'Did not know how to perform tx for {type(x)}')



################################
# * General utils
#------------------------------



scalar_types = {int, float, bool, str, Time, QuantityFloat, QuantityInt, ZefEnumValue}

def make_enum_aet(x):
    """ hacky work around function for now:
    e.g. given an enum value "EN.Color.White"
    we want to convert to the AET type expression: "AET.Enum.Color"
    """
    enum_typename: str = x.enum_type
    return getattr(AET.Enum, enum_typename)

def make_qf_aet(x):
    """ hacky work around function for now:
    e.g. given an enum value "QuantityFloat(2.1, EN.Unit.kilogram)"
    we want to convert to the AET type expression: "AET.QuantityFloat.kilogram"
    """
    quantity_unit: str = x.unit.enum_value
    return getattr(AET.QuantityFloat, quantity_unit)

def make_qi_aet(x):
    """ hacky work around function for now:
    e.g. given an enum value "QuantityInt(2.1, EN.Unit.kilogram)"
    we want to convert to the AET type expression: "AET.QuantityInt.kilogram"
    """
    quantity_unit: str = x.unit.enum_value
    return getattr(AET.QuantityInt, quantity_unit)

map_scalar_to_aet_type = {
    int:                lambda x: AET.Int,
    float:              lambda x: AET.Float,
    bool:               lambda x: AET.Bool,
    str:                lambda x: AET.String,
    Time:               lambda x: AET.Time,
    ZefEnumValue:       make_enum_aet,
    QuantityFloat:      make_qf_aet, 
    QuantityInt:        make_qi_aet, 
    }

def get_curried_arg(op, n):
    """ 
    utility function to get 
    e.g. get_curried_arg(Z['id1'][42], 0)       # => 'id1'
    e.g. get_curried_arg(Z['id1'][42], 1)       # => 42
    """
    # return op.el_ops[arg_indx][1][n]
    return LazyValue(op) | absorbed | nth[n] | collect



def most_recent_rae_on_graph(origin_uid: str, g: Graph)->ZefRef:
    """
    Will definitely not return a BT.ForeignInstance, always an instance.
    It could be that the node asked for has its origin on this graph 
    (the original rae may still be alive or it may be terminated)

    Args:
        origin_uid (str): the uid of the origin rae we are looking for
        g (Graph): on which graph are we looking?

    Returns:
        ZefRef: this graph knows about this: found instance
        None: this graph knows nothing about this RAE
    """
    if origin_uid not in g:
        return None     # this graph never knew about a RAE with this origin uid

    zz = g[origin_uid]
    if BT(zz) in {BT.FOREIGN_ENTITY_NODE, BT.FOREIGN_ATOMIC_ENTITY_NODE, BT.FOREIGN_RELATION_EDGE}:
        from .graph_slice import get_instance_rae
        return get_instance_rae(origin_uid, now(g))
        
    elif BT(zz) in {BT.ENTITY_NODE, BT.ATOMIC_ENTITY_NODE, BT.RELATION_EDGE}:
        if zz | exists_at[now(g)] | collect:
            return zz | in_frame[now(g)] | collect
        else:
            return None
    else:
        raise RuntimeError("Unexpected option in most_recent_rae_on_graph")
        



# Generation of IDs and being able to detect whether an ID was auto-generated.

generated_prefix = "tmp_id_"
def make_generate_id():
    auto_id_counter = 0
    def generate_id():
        """count up"""
        nonlocal auto_id_counter
        auto_id_counter += 1
        return f"{generated_prefix}{auto_id_counter}"
    return generate_id

def is_generated_id(x):
    if not isinstance(x, str):
        return False
    return x.startswith(generated_prefix)

def filter_temporary_ids(receipt):
    return (receipt
            | items
            | filter[first | Not[is_generated_id]]
            | func[dict]
            | collect)



@func
def single_or(itr, default):
    itr = iter(itr)
    try:
        ret = next(itr)
        try:
            next(itr)
            raise Exception("single_or detected more than one item in iterator")
        except StopIteration:
            return ret
    except StopIteration:
        return default

@func
def get_absorbed_id(obj):
    # THIS SHOULDN"T BE NEEDED! FIX!
    if is_a(obj, RT) or is_a(obj, ZefOp):
        obj = LazyValue(obj)
    return obj | absorbed | single_or[None] | collect


@func
def merge_no_overwrite(a,b):
    d = {**a}

    for k,v in b.items():
        if k in d and d[k] != v:
            raise Exception("The internal id '{k}' refers to multiple objects, including '{d[k]}' and '{v}'. This is ambiguous and not allowed.")
        d[k] = v
    return d



################################################
# * Old not used anymore?
#----------------------------------------------



def verify_assign_values(assign_v, id_definitions):
    new_value = assign_v['new_value']

    # This function is called from 2 different locations. First inside the AssignValue_ constructor where we are able to cast a ZefRef to its rae_type
    # Secondly, inside the GraphDelta constructor where we check against uid inside the id_definitions dict.
    if "zr_type" in assign_v:
        type_of_assignedto = assign_v['zr_type']
    else:
        if assign_v['some_id'] in id_definitions: type_of_assignedto = id_definitions[assign_v['some_id']]
        else: type_of_assignedto = None

    # This comes first as _eq_ for ZefRef throws in the next if statement
    if isinstance(new_value, ZefRef) or isinstance(new_value, EZefRef) or isinstance(new_value, ZefRefs) or isinstance(new_value, EZefRefs):
        raise RuntimeError(f'The passed value can\'t be of type Zefref(s) or EZefRef(s)')

    # If no value was passed
    if new_value == None:
        raise RuntimeError(f'{assign_v} is missing a value to be assigned.')

    # If type of the value passed isn't part of this set of allowed types
    if type(new_value) in scalar_types:
        raise RuntimeError(f'The type of the value passed in {assign_v} isn\'t of the allowed types: str, int, float, bool, Time, QuantityFloat, QuantityInt, ZefEnumValue')

    if type_of_assignedto: 
        type_map = {
            str:           AET.String,
            ZefEnumValue:  AET.Enum,
            QuantityFloat: AET.QuantityFloat, 
            QuantityInt:   AET.QuantityInt, 
            Time:          AET.Time,
        }
        # If type of the rae that is being assigned isn't of an AET type i.e assigning a str to an ET.Cat
        if type_of_assignedto not in {AET.Int, AET.Float, AET.Bool, AET.String, AET.Enum, AET.QuantityFloat, AET.QuantityInt, AET.Time}:
            raise RuntimeError(f'The type_of_assignedto={type_of_assignedto} must be of the allowed types: AET.Int, AET.Float, AET.Bool, AET.String, AET.Enum, AET.QuantityFloat, AET.QuantityInt, AET.Time')

        # Verify cast-ability
        elif type(new_value) == int:
            if type_of_assignedto not in {AET.Int, AET.Float, AET.Bool}:
                raise RuntimeError(f'Can only assign values of type int to only AET.Int, AET.Float, AET.Bool')
            if type_of_assignedto == AET.Bool and new_value not in [0,1]:
                raise RuntimeError(f'Can\'t assign int value that isn\'t 0 or 1 to AET.Bool')
            
        elif type(new_value) == float:
            if type_of_assignedto not in {AET.Int, AET.Float}:
                raise RuntimeError(f'Can only assign values of type float to only AET.Int, AET.Float')
            import math
            if type_of_assignedto == AET.Int and math.fabs(new_value - round(new_value)) > 1e-8:
                raise RuntimeError(f'Can only assign values of type float to int if the double is numerically sufficiently close to make rounding safe.')

        elif type(new_value) == bool:    
            if type_of_assignedto not in {AET.Int, AET.Bool}:
                raise RuntimeError(f'Can only assign values of type bool to only AET.Int, AET.Bool')
            if type_of_assignedto == AET.Int and new_value not in [False, True]:
                raise RuntimeError(f'Can\'t assign bool value that isn\'t True or False to AET.Int')

        # These are only castable one-to-one
        elif type_map[type(new_value)] != type_of_assignedto:
            raise RuntimeError(f'Can\'t assign value of type {type(new_value)} to an AET of type {type_of_assignedto}')









# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "grant",
    "revoke",
    "login",
    "logout",
]

from ..core import *
from ..core._ops import *

def privileges_implementation(user, action, privilege, target):
    # This just creates the effect, the actual implementation of the privileges
    # is in the fx module.

    if privilege == append:
        privilege = KW.append
        
    opts = {
        (KW.grant, KW.view): FX.Privileges.GrantView,
        (KW.grant, KW.append): FX.Privileges.GrantAppend,
        (KW.grant, KW.host): FX.Privileges.GrantHost,
        (KW.grant, KW.discover): FX.Privileges.GrantDiscover,
        (KW.grant, KW.modify_rights): FX.Privileges.GrantModifyRights,
        (KW.revoke, KW.view): FX.Privileges.RevokeView,
        (KW.revoke, KW.append): FX.Privileges.RevokeAppend,
        (KW.revoke, KW.host): FX.Privileges.RevokeHost,
        (KW.revoke, KW.discover): FX.Privileges.RevokeDiscover,
        (KW.revoke, KW.modify_rights): FX.Privileges.RevokeModifyRights,
    }
    typ = opts.get((action,privilege), None)
    if typ is None:
        raise Exception(f"Unknown (action,privilege) combination {(action,privilege)}")

    return Effect({
        "type": typ,
        "user": user,
        "target": target
    })

def privileges_type_info(op, curr_type):
    return VT.Effect
from ..core.op_implementations.dispatch_dictionary import _op_to_functions
_op_to_functions[RT.Privileges] = (privileges_implementation, privileges_type_info)

privileges = make_zefop(RT.Privileges)

grant = privileges[KW.grant]
revoke = privileges[KW.revoke]

login = Effect({"type": FX.ZefHub.Login})
logout = Effect({"type": FX.ZefHub.Logout})
# We don't want complaints when these are garbage collected
login._been_run = True
logout._been_run = True



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import _version
__version__ = _version.get_versions()['version']

########################################################
# * Exposing common functions
#------------------------------------------------------

# This set of imports is to define the order. Later imports are the ones to
# actually provide useful exports.
from . import core
from . import pyzef
from . import ops

from .core import *

pyzef.internals.finished_loading_python_core()

import os
if os.environ.get("ZEFDB_DEVELOPER_CIRCULAR_IMPORTS", "FALSE") == "TRUE":
    from .circular_imports import disable_check_circular_imports
    disable_check_circular_imports()

############################################
# * Starting the butler
#------------------------------------------

# Putting this into a function to avoid polluting namespace
def _autostart_behaviour():
    auto_start = ops.config("butler.autoStart", KW.get)
    if auto_start == "true":
        core.internals.initialise_butler()

_autostart_behaviour()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *

class MyTestCase(unittest.TestCase):

    def test_attach(self):
        g = Graph()

        z = instantiate(ET.Machine, g) | fill_or_attach[RT.Weight, QuantityFloat(100.0, EN.Unit.kilogram)] | collect
        z2 = (instantiate(ET.Machine, g)
              | fill_or_attach[RT.Name, AET.String, "asdf"]
              | fill_or_attach[RT.Status, "asdf"]
              | collect)
        z3 = instantiate(ET.Machine, g) | fill_or_attach[[(RT.Number, 5),
                                                         (RT.Time, now()),
                                                         (RT.Fraction, 0.5),
                                                         (RT.Status, EN.Status.On),
                                                         (RT.Disabled, True)]] | collect

        self.assertEqual(length(z | out_rels[RT]), 1)
        self.assertEqual(length(z2 | out_rels[RT]), 2)
        self.assertEqual(length(z3 | out_rels[RT]), 5)


    def test_shorthand(self):
        g = Graph()

        z = ET.Dog | g | run
        self.assertTrue(is_a(z, ET.Dog))

        z1, [(name_rel, name),
             (breed_rel, breed)] = (z, [(RT.Name, "Rufus"),
                                        (RT.Breed, "Labrador")]) | g | run

        self.assertEqual(to_ezefref(z), to_ezefref(z1))
        self.assertTrue(is_a(name_rel, RT.Name))
        self.assertEqual(value(name), "Rufus")
        self.assertTrue(is_a(breed_rel, RT.Breed))
        self.assertEqual(value(breed), "Labrador")

        z1, none, [a,b,c] = (z, RT.Nickname, ["Dog", "Boy", "Mongrel"]) | g | run

        self.assertEqual(to_ezefref(z), to_ezefref(z1))
        self.assertEqual(none, None)
        self.assertEqual(value(a), "Dog")
        self.assertEqual(value(b), "Boy")
        self.assertEqual(value(c), "Mongrel")
        self.assertTrue(has_relation(now(z), RT.Nickname, a))
        self.assertTrue(has_relation(now(z), RT.Nickname, b))
        self.assertTrue(has_relation(now(z), RT.Nickname, c))

        [a,b], none, z1 = (["Alice", "Bob"], RT.Owner, z) | g | run

        self.assertEqual(to_ezefref(z), to_ezefref(z1))
        self.assertEqual(none, None)
        self.assertEqual(value(a), "Alice")
        self.assertEqual(value(b), "Bob")
        self.assertTrue(has_relation(a, RT.Owner, now(z)))
        self.assertTrue(has_relation(b, RT.Owner, now(z)))

    def test_shorthand_assign(self):
        g = Graph()

        ae = AET.String | g | run
        self.assertEqual(value(ae), None)

        ae | assign_value["word"] | g | run
        self.assertEqual(value(now(ae)), "word")

        g | now | all[AET] | single | assign_value["single"] | g | run
        self.assertEqual(value(now(ae)), "single")

if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from zef import *
from zef.ops import *

import asyncio


time_before_parent_sleep = None
time_after_parent_sleep = None

time_inside_client = None

async def Parent(g):
    global time_before_parent_sleep
    global time_after_parent_sleep

    with Transaction(g):
        time_before_parent_sleep = now()
        await asyncio.gather(asyncio.sleep(5), Child(g))
        time_after_parent_sleep = now()

async def Child(g):
    global time_inside_client
    await asyncio.sleep(1)
    with Transaction(g):
        time_inside_client = now()
    
class MyTestCase(unittest.TestCase):
    def test_no_trans(self):
        g = Graph()
        # TODO: Need to fix this!
        with self.assertRaises(Exception):
            asyncio.run(Parent(g))

        # This is for when things can work, or at least wait on one another.
        #self.assertTrue(time_before_parent_sleep < time_inside_client < time_after_parent_sleep)

if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *


class MyTestCase(unittest.TestCase):
    def test_zef_enum_value(self):
        g = Graph()
        with Transaction(g):
            enum_type = "Zeyad"
            enum_val = "TEST"
            EN(enum_type, enum_val)
            ae = AET.Enum(enum_type) | g | run
            (ae <= EN(enum_type, enum_val)) | g | run

        retrieved_value = ae | value | collect
        self.assertEqual(retrieved_value.enum_value, enum_val)
        self.assertEqual(retrieved_value.enum_type, enum_type)


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import networkx
from .. import *
from ..ops import *

from enum import Enum
class Direction(Enum):
    BOTH = 1
    OUTGOING = 2
    INCOMING = 3

# This class needs to tie in with ZefRef_to_node_ref
class NodeRef:
    def __init__(self, z):
        assert isinstance(z, ZefRef)
        self.z = z

    def __repr__(self):
        return f"Node(#{index(self.z)})"
    def __eq__(self, other):
        if not isinstance(other, NodeRef):
            return NotImplemented
        return self.z == other.z
    def __hash__(self):
        return hash(self.z)

# This class needs to tie in with ZefRef_to_edge_ref
# class EdgeRef:
#     def __init__(self, z):
#         assert isinstance(z, ZefRef)
#         self.z = z

#     def __repr__(self):
#         src = source(self.z)
#         trg = target(self.z)
#         return f"Edge(#{index(src)} -> #{index(trg)})"

#     def __eq__(self, other):
#         if not isinstance(other, EdgeRef):
#             return NotImplemented
#         return self.z == other.z

#     def __hash__(self):
#         return hash(self.z)

@func
def ZefRef_to_node_ref(z):
    # Here we provide a consistent "name" for a ZefRef to library functions that
    # use the ProxyGraph. Even though we allow users to provide pretty much
    # anything that fits into g[x], this consistency here should provide a
    # better fit for algorithms that do comparisons, etc...
    # return index(z)
    return NodeRef(z)

@func
def ZefRef_to_edge_ref(z, direction):
    src_ref = ZefRef_to_node_ref(source(z))
    trg_ref = ZefRef_to_node_ref(target(z))
    if direction == Direction.INCOMING:
        return (trg_ref, src_ref)
    elif direction == Direction.OUTGOING:
        return (src_ref, trg_ref)
    else:
        assert direction == Direction.BOTH
        # For consistency, make the smallest index the first node
        if index(source(z)) < index(target(z)):
            return (src_ref, trg_ref)
        else:
            return (trg_ref, src_ref)
    # return EdgeRef(z)

@func
def all_edges(z, rt, direction):
    return all_edges_with_end(z, rt, direction) | map[first]

@func
def all_edges_with_end(z, rt, direction):
    if direction == Direction.BOTH:
        return concat(all_edges_with_end(z, rt, Direction.OUTGOING),
                      all_edges_with_end(z, rt, Direction.INCOMING))
    elif direction == Direction.OUTGOING:
        return z | out_rels[rt] | map[lambda z: (z, target(z))] | collect
    else:
        assert direction == Direction.INCOMING
        return z | in_rels[rt] | map[lambda z: (z, source(z))] | collect

class ProxyGraph:
    def __init__(self,
                 gs: GraphSlice,
                 node_filter=ET, # Can be a list (explicit list of ZefRefs) or something that is_a accepts.
                 rts=RT, # must be something that is_a accepts, returning relations
                 *args,
                 undirected=False,
                 multigraph=False,
                 reversed=False,
                 include_type_as_field=True,
                 **kwds
                 ):
        if isinstance(gs, Graph):
            raise Exception("Create a ProxyGraph with a GraphSlice instead of a Graph. Use `now(g)` if you want the latest slice.")
        if not isinstance(gs, GraphSlice) or args or kwds:
            raise Exception("Trying to create a graph with args/kwds we don't understand. If you are encountering this error as a result of running a networkX algorithm, it is likely that the algorithm is trying to construct a new graph of the same type. Construction of a graph is not possible with this class. You can instaed pass a native networkX class to the algorithm by writing, for example, `nx.minimum_spanning_tree(pg.to_native())`")

        self.gs = gs
        self.node_filter = node_filter
        self.rts = rts
        self.undirected = undirected
        self.multigraph = multigraph
        self.reversed = reversed
        self.include_type_as_field = include_type_as_field
        assert not multigraph, "Currently no multigraph support"

        # On creation, we make a consistent mapping of indices, just so we can
        # be sure this won't change in the future.
        self.node_mapping = self._all_node_refs()

        if self.undirected:
            self.direction = Direction.BOTH
        elif self.reversed:
            self.direction = Direction.INCOMING
        else:
            self.direction = Direction.OUTGOING
        
    def _all_node_refs(self):
        if isinstance(self.node_filter, tuple) or isinstance(self.node_filter, ZefRefs):
            raise Exception("Don't pass a tuple or a ZefRefs to ProxyGraph as the node_filter")

        if isinstance(self.node_filter, list):
            if isinstance(self.node_filter[0], NodeRef):
                return self.node_filter
            if isinstance(self.node_filter[0], ZefRef):
                return [ZefRef_to_node_ref(x) for x in self.node_filter]
            raise TypeError("Can't understand node_filter")
        else:
            return self.gs | all[self.node_filter] | func[list] | map[ZefRef_to_node_ref] | collect

    @property
    def nodes(self):
        return ProxyNodeView(self)

    @property
    def edges(self):
        return ProxyEdgeView(self)

    @property
    def adj(self):
        return ProxyAdjacencyView(self)

    @property
    def succ(self):
        return ProxyAdjacencyView(self)

    @property
    def pred(self):
        return ProxyAdjacencyView(self.reverse())

    def successors(self, n):
        return iter(self.pred[n])
    neighbors = successors

    def predecessors(self, n):
        return iter(self.succ[n])

    @property
    def degree(self):
        return ProxyDegreeView(self, Direction.BOTH)

    @property
    def out_degree(self):
        if self.reversed:
            return ProxyDegreeView(self, Direction.INCOMING)
        else:
            return ProxyDegreeView(self, Direction.OUTGOING)

    @property
    def out_degree(self):
        if self.reversed:
            return ProxyDegreeView(self, Direction.OUTGOING)
        else:
            return ProxyDegreeView(self, Direction.INCOMING)

    def __repr__(self):
        info = []
        if self.reversed:
            info += ["reversed"]
        if isinstance(self.node_filter, list):
            info += [f"subgraph of {len(self.node_filter)} nodes"]
        elif self.node_filter != ET:
            info += [f"ET={self.node_filter}"]
        if self.rts != RT:
            info += [f"RT={self.rts}"]
        if self.multigraph:
            info += [f"multigraph"]
        if self.multigraph:
            info += [f"undirected"]
        info += [f"{repr(self.gs)}"]
        return "ProxyGraph(" + ", ".join(info) + ")"
    def _repr_pretty_(self, p, cycle):
        p.text(str(self))

    def number_of_nodes(self):
        return len(self.nodes)
    def number_of_edges(self, u, v):
        raise NotImplementedError("Need this for multigraphs")
        return len(self.edges)

    def __getitem__(self, item):
        # TODO: Some of this should be moved to e.g. g.nodes[1] logic


        # Returns a view from a node which is the first half of an edge query.
        resolved = None
        if isinstance(item, NodeRef):
            resolved = item.z
        elif isinstance(item, str) or is_a(item, uid) or is_a(item, int):
            resolved = self.gs[item]
        elif isinstance(item, ZefRef):
            resolved = in_frame(item, self.gs)

        if not resolved:
            raise Exception(f"Couldn't find node {item}")

        return ProxyAtlasView(resolved, self)

    def __iter__(self):
        return iter(self.node_mapping)
    def __len__(self):
        return len(self.node_mapping)
    def nbunch_iter(self, nbunch=None):
        if nbunch is None:
            return iter(self)
        nbunch = [ZefRef_to_node_ref(self[x].z) for x in nbunch]
        return iter(set(nbunch) & set(self.node_mapping))
        

    # Graph properties
    def is_directed(self):
        return not self.undirected
    def is_multigraph(self):
        return self.multigraph


    # TODO:
    # has_edge
    # adjacency
    def order(self):
        return self.number_of_nodes()
    # size
    # subgraph - this could be done with an explicit node list, rather than ets
    def subgraph(self, nodes):
        return ProxyGraph(self.gs,
                           nodes,
                           self.rts,
                           undirected=self.undirected,
                           multigraph=self.multigraph,
                           reversed=self.reversed,
                           include_type_as_field=self.include_type_as_field,
                          )
                           
    def reverse(self):
        return ProxyGraph(self.gs,
                           self.node_mapping,
                           self.rts,
                           undirected=self.undirected,
                           multigraph=self.multigraph,
                           reversed=not self.reversed,
                           include_type_as_field=self.include_type_as_field,
                          )

    def to_undirected(self):
        return ProxyGraph(self.gs,
                           self.node_mapping,
                           self.rts,
                           undirected=True,
                           multigraph=self.multigraph,
                           reversed=self.reversed,
                           include_type_as_field=self.include_type_as_field,
                          )

    def to_directed(self):
        return ProxyGraph(self.gs,
                           self.node_mapping,
                           self.rts,
                           undirected=False,
                           multigraph=self.multigraph,
                           reversed=self.reversed,
                           include_type_as_field=self.include_type_as_field,
                          )

    def to_native(self, include_fields=False):
        import networkx as nx
        assert not self.multigraph
        if self.undirected:
            nxg = nx.Graph()
        else:
            nxg = nx.DiGraph()

        if include_fields:
            nxg.add_nodes_from(self.nodes.items())
            nxg.add_edges_from(self.edges.data())
        else:
            nxg.add_nodes_from(self.nodes)
            nxg.add_edges_from(self.edges)

        return nxg
            

# Note: this does not represent the node itself (that is whatever ZefRef_to_ref
# returns) but this is rather the interface that allows modification of the
# node's properties smoothly.
class ProxyNodeView:
    def __init__(self, dg):
        self.dg = dg

    # def __repr__(self):
    #     return f"Node({repr(ET(self.z))}:{self._to_ref()})"
    # def _repr_pretty_(self, p, cycle):
    #     p.text(str(self))

    def __getitem__(self, node):
        z = self.dg[node].z
        return get_props_on(z, include_type=self.dg.include_type_as_field)
    
    # def __setitem__(self, item, val):
    #     # Set props
    #     raise NotImplementedError()

    def __iter__(self):
        return iter(self.dg.node_mapping)
    def __len__(self):
        return len(self.dg.node_mapping)

    def items(self):
        for n in self:
            yield n,self[n]

    def data(self):
        for n in self:
            yield n,self[n]

    def __call__(self, data=False, default=None):
        assert data is False
        return self
        

class ProxyEdgeView:
    def __init__(self, dg, dataview=False):
        self.dg = dg
        self.dataview = dataview

    # def __repr__(self):
    #     return f"Edge({repr(RT(self.z))}:{self._to_ref()})"
    # def _repr_pretty_(self, p, cycle):
    #     p.text(str(self))

    def __getitem__(self, e):
        if isinstance(e,tuple):
            src,trg = e
            z_src = self.dg[src].z
            z_trg = self.dg[trg].z
            rel = (#z_src > L[self.dg.rts]
                z_src | all_edges_with_end[self.dg.rts][self.dg.direction]
                | filter[second | equals[z_trg]]
                | map[first]
                | single
                | collect)
            return get_props_on(rel, include_type=self.dg.include_type_as_field)
        else:
            # This is a Zef-specific extension.

            # This is a single-arg call, it should be the ZefRef of the relation directly
            rel = e
            src = source(rel)
            trg = target(rel)
            assert is_a(rel, self.dg.rts)
            # Throw exceptions if the src/trg is not in the valid node list
            self.dg[src]
            self.dg[trg]
            return get_props_on(rel, include_type=self.dg.include_type_as_field)
    
    # def __setitem__(self, item, val):
    #     # Set props
    #     raise NotImplementedError()

    def __iter__(self):
        if self.dg.undirected:
            # When grabbing all edges, we don't want to double up.
            direction_override = Direction.OUTGOING
        else:
            direction_override = self.dg.direction
        out = (self.dg.node_mapping
                 | map[lambda x: x.z]
                 | map[all_edges_with_end[self.dg.rts][direction_override]]
                 | concat
                 | filter[second | func[ZefRef_to_node_ref] | contained_in[self.dg.node_mapping]]
                 | map[first]
                 | map[ZefRef_to_edge_ref[self.dg.direction]])
        if self.dataview:
            out = out | map[lambda x: (*x, self.__getitem__(*x))]
        return iter(out)
    def __len__(self):
        return length(iter(self))
    def __contains__(self, e):
        # The logic here is only unusual for undirected graphs, where the
        # ordering of the edges shouldn't matter. For consistency, we have made
        # the node with the smallest index on the graph be the first.
        # u,v = e
        # if self.dg.undirected:
        #     if index(u.z) > index(v.z):
        #         u,v = v,u
        # return (u,v) in iter(self)

        # I don't like this, but it's the fastest for now
        try:
            u,v = e
            self[u,v]
            return True
        except StopIteration:
            # TODO: This might change in the future with zef, so will have to update this
            return False

    def items(self):
        for e in self:
            yield n,self[n]

    def data(self):
        for e in self:
            yield e[0],e[1],self[e]

    def __call__(self, data=False, default=None):
        assert default is None
        return ProxyEdgeView(self.dg, data)
            

def get_props_on(z, include_type):
    props = {}
    for rel in z | out_rels[RT] | filter[target | is_a[AET]]:
        props[str(RT(rel))] = rel|target|value|collect
    if include_type:
        props["type"] = rae_type(z)
    return props

# This is the view of a node in the context of a ProxyGraph, that allows for
# querying of its edges. In other words, it is the first half of an edge query,
# completed by a getitem or similar on this object.
class ProxyAtlasView:
    def __init__(self, z : ZefRef, dg : ProxyGraph):
        self.z = z
        self.dg = dg

    def __repr__(self):
        return f"AtlasView({ZefRef_to_node_ref(self.z)})"
    def _repr_pretty_(self, p, cycle):
        p.text(str(self))

    def __iter__(self):
        # This returns the neighbors, not the edges
        # all_edges_with_end returns a tuple with (rel,ent)
        lazy = self.z | all_edges_with_end[self.dg.rts][self.dg.direction] | map[second] | map[ZefRef_to_node_ref] | filter[contained_in[self.dg.node_mapping]]
        return iter(lazy)
    def __len__(self):
        return length(iter(self))

    def __getitem__(self, other):
        # The other is a node itself, so we need to look it up and then find the
        # relation between them.

        # We can work with another ProxyAtlastView ourselves
        other_view = self.dg[other]
        other_z = other_view.z

        # all_edges_with_end returns a tuple with (rel,ent)
        opts = self.z | all_edges_with_end[self.dg.rts][self.dg.direction] | filter[second | equals[other_z]] | map[first] | collect
        if len(opts) == 0:
            raise IndexError(f"There is no edge between objects {ZefRef_to_zef(self.z)} and {ZefRef_to_zef(other_z)}")

        rel = single(opts)

        return get_props_on(rel, include_type=self.dg.include_type_as_field)

# I am a little confused how repetitive this implementation seems...
class ProxyAdjacencyView:
    def __init__(self, dg):
        self.dg = dg

    def __iter__(self):
        return iter(self.dg.nodes)
    def __len__(self):
        return len(self.dg.nodes)

    def __getitem__(self, item):
        return self.dg[item]

    def items(self):
        return ((x, self[x]) for x in iter(self))

class ProxyDegreeView:
    def __init__(self, dg, direction, nbunch=None):
        self.dg = dg
        self.direction = direction
        self.nbunch = nbunch

    def __call__(self, nbunch=None):
        if nbunch is None:
            return self
        if isinstance(nbunch, list):
            return ProxyDegreeView(self.dg, self.direction, nbunch)

        return self[nbunch]

    def __getitem__(self, n):
        return self.dg[n].z | all_edges[self.dg.rts][self.direction] | length | collect

    def __iter__(self):
        lazy = LazyValue(iter(self.dg))
        if self.nbunch is not None:
            lazy = lazy | filter[contained_in[self.nbunch]]
        lazy = lazy | map[lambda x: (x,self[x])]
        return iter(lazy)

    def __len__(self):
        return len(self.dg)



##################################################################
# * Old ideas for writing to graph
#----------------------------------------------------------------

    # def add_node(self, hashable, *, typ=ET.Node, **kwds):
    #     h = hash(hashable)
    #     import ctypes
    #     h1 = ctypes.c_int32(h >> 32).value
    #     h2 = ctypes.c_int32(h & 0xFFFFFFFF).value
    #     if hash(hashable) in self.node_lookup:
    #         return self.node_lookup[h]

    #     z = instantiate(typ, self.g) | attach[[
    #         (RT._Hash1, h1),
    #         (RT._Hash2, h2),
    #         *( (RT(key), val) for key,val in kwds),
    #     ]]

    #     self.object_lookup[h] = hashable
    #     self.node_lookup[h] = z | to_ezefref
    #     return z | to_ezefref

    # def add_nodes_from(self, hashables, *, typ=ET.Node, **kwds):
    #     # TODO: Handle adding from another graph

    #     for hashable in hashables:
    #         self.add_node(self, hashable, typ=typ, **kwds)

    # def add_edge(self, na, nb, *, typ=RT.Edge, **kwds):
    #     z_a = self.add_node(na)
    #     z_b = self.add_node(nb)

    #     if has_relation(z_a|now, typ, z_b|now):
    #         return relation(z_a|now, typ, z_b|now)

    #     z = instantiate(z_a, typ, z_b, self.g)
    #     return z



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..pyzef.main import (
    EntityType,
    Graph,
    Keyword,
    QuantityFloat,
    QuantityInt,
    RelationType,
    Time,
    TimeSlice,
    EZefRef,
    EZefRefs,
    EZefRefss,
    ZefEnumValue,
    ZefRef,
    ZefRefs,
    ZefRefss,
    Zwitch,
    blobs,
    currently_open_tx,
    data_layout_version,
    days,
    hours,
    index,
    instantiate,
    lookup_uid,
    make_primary,
    merge,
    minutes,
    months,
    revision_graph,
    seconds,
    set_keep_alive,
    weeks,
    years,
    zearch,
    zef_get,
    zwitch
)

from ..pyzef.internals import (
    ET,
    RT,
    EN,
    AET,
    BT,
    AtomicEntityType,
    KW,
    Delegate,
)

from .internals import Transaction


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
TODO: make the decorator @zef_function(g, ...) also work if we don't have the primary instance: remote merge if someone else has the primary instance

z_fct can be a ET.ZEF_Function
inputs and outputs should be plain data: elementary python or ZefTensors, / flat graphs
If python Zef Functions are type annotated, the zef runtime can be used to enforce type checking (optional setting)

How does one define a zef function? Use a decorator @zef_function(g, label='square function', f1, f2) above a function definition
What does this do?
A)   side effect: extracts the functions source code, specified meta data, closure over other 
     functions (should we allow variables? e.g. plain data like lists, also ZefRefs) and saves all of this on the specified graph.
B)   What should the decoration process return? 
     Firstly, something that behaves like the function that was decorated. This could be the actual function, 
     but we may want to close over this very zef function in other zef functions. These need to know 

why do we want to overload the __call__ method for a ZefRef? Then the syntax becomes easier: we can just do z_my_zef_fct(42, 'x')

Future options: provide optional specification of exception behavior as well?

-------------------- Example -----------------

@zef_function(
    g,
    graph_spec = z_some_spec,   # specs are just boolean functions

    my_fct = z_fct5,            # binding of other zef functions to names
    my_fct2 = z_fct45,     
)
def my_fct(x: z_spec1,y: z_spec2, z: z_spec3) -> z_spec4:
    return 42

"""





from typing import Optional, Union
import typing
import inspect
import pyperclip
import traceback

from . import internals
from ._core import *
from ._ops import *


# IMPORTANT!!!
# IMPORTANT!!!
# IMPORTANT!!!
# NOTE: This definition of func must live here, separated from the ops package.
# This is to break any dependencies other modules (e.g. the fx package) have on
# func, without themselves depending on ops and causing a cyclic dependency
# chain.
# IMPORTANT!!!
# IMPORTANT!!!

##############################
# * Func op
#----------------------------

# This provides the zefop-like behaviour expected from a func. Needs to support theset hings:
# - z | func[f] | run
#
# - @func
#   def f(x, y):
#     ...
#   z | f[42] | run
#
# - g = Graph()
#   @func(g)
#   def f(x, y):
#     ...
#   z | f[42] | run
#
# - z | func[lambda x: x] | run



# This is not a zefop, it only becomes one when [] is applied to it.
class FunctionConstructor:
    """
    1. -------------------------------------------------------
     func[lambda x: 2*x]     # close over Python function
    
    2. -------------------------------------------------------
    func[z_my_function_zefref]      # why may we want this rather than func(z_my_function_zefref)? Possibly useful if this appears in zefop chain. (?) Unsure 
    
    3. -------------------------------------------------------
    func[z_my_function_ezefref]    # Zef Functions are values, any time slice defines the same one

    4. -------------------------------------------------------
    def ff(x, y):
        return x+2*y

    func(ff)            # close over Python function
    
    
    5. -------------------------------------------------------
    @func
    def ff(x, y):
        return x+2*y

        
    6. -------------------------------------------------------
    @func(g)
    def ff(x, y):
        return x+2*y

    """
    @staticmethod
    def __call__(*args, **kwds):
        from types import FunctionType
        from .abstract_raes import Entity
        if len(kwds) == 0 and len(args) == 1 and isinstance(args[0], FunctionType):
            return ZefOp(((RT.Function, ((1, args[0]), )), ))
        else:
            from zef.core.zef_functions import zef_function_decorator, _local_compiled_zef_functions, time_resolved_hashable
            promote_to_zefref_func = zef_function_decorator(*args, **kwds)
            def inner(func):
                zefref = promote_to_zefref_func(func)
                abstract_entity = Entity(zefref)
                _local_compiled_zef_functions[abstract_entity.d['uid']] = _local_compiled_zef_functions[time_resolved_hashable(zefref)]
                return ZefOp(((RT.Function, ((0, abstract_entity), )), ))
            return inner

    @staticmethod
    def __getitem__(arg):
        # TODO we gotta check if arg is of type Zef Lambda once we implement it
        # return ZefOp(((RT.Function, ((1, arg), )), ))
        return ZefOp(((RT.Function, ((1, arg), )), ))


func = FunctionConstructor()

##################################
# * Implementation
#--------------------------------





_local_compiled_zef_functions = {}
# do not realease the graphs from which zef functions are read to prevent garbage collection. We may need to access meta info
# _graphs_held_on_to = set()



def time_resolved_hashable(z_fct):
    """we can't use zefrefs as keys in a dict: z1==z2 
    comparison was disabled in zefdb for safety. Make a tuple of uzefrefs."""
    return (z_fct | to_ezefref | collect, z_fct | frame | collect)



def compile_in_zef_context(fct_str: str, additional_globals: dict, fake_filename="__zef__"):
    from .. import core
    from .. import ops
    shared_globals = {
        **core.__dict__,
        **ops.__dict__,
        **additional_globals,     # convert the list of tuples to a dict: the first tuple element is the fct name as a str
        } # pass variables in, but compiled fct will be in there too
    code = compile(fct_str, fake_filename, "exec")
    # if type annotations (both input and output) are made, these show up as the 
    # first elements in code.co_names tuple
    # e.g. for def mapper(x: str) -> int: 
    # code.co_names=('str', 'int', 'zef_function_4718c414b5be00ce70b2d59bbda46a2bba8610950c168e78')
    assert len(code.co_names) >= 1, "Function to compile produced multiple top-level names"
    exec(code, shared_globals)     # don't give the fct direct access to our globals here
    fct_name = code.co_names[-1]
    return shared_globals[fct_name]

def compile_zef_function(z_fct: ZefRef):
    """"""
    if time_resolved_hashable(z_fct) in _local_compiled_zef_functions:
        return z_fct

    g = Graph(z_fct)
    # Prevent garbage collection
    set_keep_alive(g, True)

    # first compile all zef functions bound within this zef fct's scope
    z_bound_fcts_pairs = (
        z_fct
        | out_rels[RT.Binding]
        | map[lambda z_rel: (z_rel | Out[RT.Name] | value | collect,
                             z_rel | target | in_frame[
                                 g[z_rel | Out[RT.UseTimeSlice] | value | collect]
                                 | to_graph_slice | collect
                             ] | collect)]
        | map[lambda p: (p[0], compile_zef_function(p[1]))]
        | collect
    )
    fct_str = z_fct | Out[RT.PythonSourceCode] | value | collect
    # what is in the scope of the zef function that will execute?    
    fct = compile_in_zef_context(fct_str, dict(z_bound_fcts_pairs))
    _local_compiled_zef_functions[time_resolved_hashable(z_fct)] = fct
    return fct



def zef_function_signature(z_fct):
    from inspect import signature
    from .core import rae_type, ET, ZefRef
    assert isinstance(z_fct, ZefRef)
    assert rae_type(z_fct) == ET.ZEF_Function
    
    compile_zef_function(z_fct)
    fct = _local_compiled_zef_functions[time_resolved_hashable(z_fct)]
    return signature(fct)



def is_zef_function_name(fct_name: str)->bool:
    if len(fct_name) != 61:
        return False
    if fct_name[0:13] != 'zef_function_':
        return False
    uid_maybe = fct_name[13:]
    if not set(['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f']
        ).issuperset(set(uid_maybe)):
        return False
    return True


def zef_function_decorator(g: Graph=None, label: str=None, is_pure=False, **kwargs):        
    """
    Use as a decorator above a function to beam it up into the zefnet.
    @zef_function(g, label='my favorite function')
    def my_f(x):
        ...


    - The function name given upon the very first definition is attached as z_zef_fct >> RT.OriginalName
    - if the optional label argument is specified in the decorator (a string), it is attached as z_zef_fct >> RT.Label




                                           RT.UseTimeSlice
                                       ┌──────────────────►AET.String    # hack for now until we can point at txs
                                       │
                                       │  RT.Name
                                       ├──────────────────►AET.String
                                       │
                       RT.Binding[*]   │
                    ┌──────────────────┴────────────────────────────────►ET.ZEF_Function
                    │
                    │
   ET.ZEF_Function──┤
                    │  RT.OriginalName
                    ├────────────────────►AET.String
                    │
                    │  RT.PythonSourceCode
                    ├────────────────────►AET.String
                    │
                    │  RT.Label?
                    ├────────────────────►AET.String
                    │
                    │  RT.IsPure?
                    └────────────────────►AET.Bool


    """
    assert label is None or isinstance(label, str)
    def inner(func):            
        """one more layer of indirection is needed if we want to create a decorator that takes arguments"""
        # assert isinstance(g, core.Graph)
        assert label is None or isinstance(label, str)
        # from .ops import terminate
        # make sure all kw args passed in are zef functions. We may consider binding libraries, 
        # data as parameters etc. in future if there is a compelling reason to do so.
        # For now we should try to import everything explicitly within the function.
        for fct_name, zef_fct_maybe in kwargs.items():
            if not isinstance(zef_fct_maybe, ZefRef):
                raise TypeError(f'Currently we only support binding of other zef functions within a zef function. The type passed in for the name "{fct_name}" was {type(zef_fct_maybe)}')
            if Graph(zef_fct_maybe) != g:
                raise RuntimeError(f'Currently any zef function bound to the local scope of another zef function has to live on the same graph. We are likely to relax this constraint once usage patterns become clear. Attempted to bind the zef function named "{fct_name}" {zef_fct_maybe}')


        # TODO: This is where we should analyze for call signature
        # from inspect import signature
        # print(signature(func).parameters)       # get the names of arguments of func

        if g is not None:
            return make_function_entity(g, label, is_pure, func, **kwargs)
        else:
            return LocalFunction(func, is_pure=is_pure)


    return inner

def make_function_entity(g, label, is_pure, func, **kwargs):        
    s_full = inspect.getsource(func)        
    import re
    # Doing two things here - a) finding the def line, and b) need to strip
    # common whitespace prefix from all lines.
    m = re.search(r"^(\s*)def ", s_full, re.MULTILINE)
    func_line_start = m.start()
    s = s_full[func_line_start:]

    prefix = len(m[1])
    lines = s.split('\n')
    lines = [line[prefix:] for line in lines]
    s = '\n'.join(lines)

    # what is the current function name? If it is 'zef_function_a4s564...', 
    # then we need to match it with an existing zef function. Otherwise create it new.        
    fct_name = func.__name__
    zef_function_exists = is_zef_function_name(fct_name)
    docstring_maybe = inspect.getdoc(func)

    with Transaction(g):
        if zef_function_exists:
            # zef function already exists. matching up
            zef_fct_uid = fct_name[13:]
            if zef_fct_uid not in g:
                raise RuntimeError("The uid extracted from the zef function name was not found in the graph")

            z_zef_fct = (g[zef_fct_uid] | now)
            z_python_str = z_zef_fct | Out[RT.PythonSourceCode] | collect
            # only assign a new value if the contents changed
            if (value(z_python_str)) != s:
                z_python_str | assign_value[s] | g | run

            # If docstring already exists; update it may be
            if len(z_zef_fct | Outs[RT.DocString] |collect) == 1:
                z_doctstring_str = z_zef_fct | Out[RT.DocString] | collect
                # If the Docstring was removed
                if not docstring_maybe:
                    z_zef_fct | out_rel[RT.DocString] | terminate | g | run
                # If Docstring was updated
                elif docstring_maybe != (z_doctstring_str | value | collect):
                    z_doctstring_str | assign_value[docstring_maybe] | collect
            # Case where this function existed before introducting Docstring parsing
            else:
                # If Docstring is defined attach it
                if docstring_maybe: (z_zef_fct, RT.DocString, docstring_maybe) | g | run
        else:
            # zef function does not exist yet. Make a new ET.ZEF_Function and attach the string after processing
            z_zef_fct = ET.ZEF_Function | g | run
            # replace the function name given by the user with a unique one containing the uid of the rel
            s_renamed = s[:s.find('def')] + f"def zef_function_{uid(z_zef_fct)}" + s[s.find('('):]
            (z_zef_fct, RT.PythonSourceCode, s_renamed) | g | run
            if label is not None:
                (z_zef_fct, RT.Label, label) | g | run
            if is_pure is not False:
                (z_zef_fct, RT.IsPure, is_pure) | g | run

            # attach the very first name the programmer gave as metadata (extracting 
            # the fct after this will auto-generate the name using the uid)
            (z_zef_fct, RT.OriginalName, fct_name) | g | run

            # If Docstring is defined attach it
            if docstring_maybe: (z_zef_fct, RT.DocString, docstring_maybe) | g | run


        # TODO: What if only the name of the bound variable was changed?

        # immediately focus on the diffing case, if one needs to establish the required diff of bindings present in the 
        # zef function pre and post                
        bindings_pre = dict((z_zef_fct | out_rels[RT.Binding]) 
                        | map[lambda z_rel: (z_rel | Out[RT.Name] | value, z_rel | target | to_frame[ g[z_rel | Out[RT.UseTimeSlice] | value]] )]
                        | collect
                        )
        bindings_post = kwargs
        # bindings_to_add = bindings_post - bindings_pre
        bindings_to_add = {k:v for k,v in bindings_post.items() if k not in bindings_pre}
        bindings_to_remove = {k:v for k,v in bindings_pre.items() if k not in bindings_post}

        def remove_binding(z_fct: ZefRef, name: str, z_attached_fct: EZefRef):
            z_ed = (z_fct | out_rels[RT.Binding]) | filter[lambda z_ed: z_ed | Out[RT.Name] | value == name] | only | collect
            z_ed | Out[RT.Name] | terminate | g | run
            z_ed | Out[RT.UseTimeSlice] | terminate | g | run
            z_ed | terminate | g | run

        def add_binding(z_fct: ZefRef, name: str, z_attached_fct):                    
            z_rel = (instantiate(z_fct | to_ezefref | collect, RT.Binding, z_attached_fct | to_ezefref | collect, g) 
                    | fill_or_attach[RT.Name, name] 
                    | fill_or_attach[RT.UseTimeSlice, str(base_uid(z_attached_fct | frame | to_tx))]        # the zef function passed in by the user may be pointing to an earlier 
                    | collect
                    # TODO: this should directly point at the transaction once zefDB allows attaching RTs to txs. This is just a hack for now !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                    )

        list(bindings_to_remove.items()) | for_each[lambda p: remove_binding(z_zef_fct, *p)]
        list(bindings_to_add.items()) | for_each[lambda p: add_binding(z_zef_fct, *p)]


    # Prevent garbage collection
    set_keep_alive(g, True)
    # from zefdb.tools import pipeable
    # func_pipeable = pipeable(func)
    # func_pipeable._my_zefref = z_zef_fct
    # return func_pipeable
    compile_zef_function(z_zef_fct | now | collect)     # purely runs side effects: compiles fct and adds it to _local_compiled_zef_functions. It is idempotent though.
    # to overload the function that is being called upon z(), we cannot monkey patch the actual object instance.
    # Python calls type(z).__call__(z), which we can monkey patch for all instances of the class            
    return z_zef_fct


class LocalFunction:
    def __init__(self, func, *, is_pure=False):
        self.func = func
        self.is_pure = is_pure

    def __getitem__(self, arg):
        return call[self][arg]

    def __call__(self, *args, **kwds):
        return self.func(*args, **kwds)

    def __or__(self, other):
        # Promotes the function to a ZefOp then pipes with other
        # i.e: @func
        #      def f(x): pass
        # f | map
        return call[self] | other

    def __ror__(self, other):
        # Similar to ror behavior but pipes the other way
        # i.e: map | f
        return other | call[self]
    
        


def zef_fct_to_source_code_string(z_func: ZefRef) -> str:
    from ..ops import text_art

    def _make_header(z_zef_function: ZefRef)->str:
        g = Graph(z_zef_function)         
        # TODO: possibly warn if this is NOT the latest version known at the time of generating this snippet!
        original_fct_name = z_zef_function | Out[RT.OriginalName] | value | collect if z_zef_function | has_out[RT.OriginalName] | collect else None
        original_fct_name_val = original_fct_name if original_fct_name is not None else '*unknown*'
        graph_uid = str(uid(g))
        return (
"#%% " + "-"*37 + f" zef function - original name: '{original_fct_name_val}' " + "-"*37 + "\n"
+ f"# -------------------------- this code snippet was generated at {str(now())} --------------------------\n"
+ text_art(f"---- {original_fct_name_val} ----") 
+ "\n#\n" + "\n".join([
#f"# origin: first created by {'-- ZefDB user --'} at ...",
#f"# last change: by -- ZefDB user -- ...",
# f'# z_zef_function = g["{uid(z_zef_function)}"]["{uid(z_zef_function | frame)}"]',            # TODO: update this to new uid and ref frame syntax
f"# we need to bind this to a var outside for now to prevent the graph being garbage collected.\n"
f"g_{graph_uid[:8]} = Graph('{graph_uid}')     # Graph tags: {g.tags}",
f'@func(',
f'    g = g_{graph_uid[:8]},',
f"    label = '{value(z_zef_function | Out[RT.Label])}'," if z_zef_function | has_out[RT.Label] | collect else None,
f"    is_pure = {value(z_zef_function | Out[RT.IsPure])}," if z_zef_function | has_out[RT.IsPure] | collect else None,
*[f"    {value(z_ed | Out[RT.Name])} = g['{uid(z_ed | target)}'] | to_frame[g['{value(z_ed | Out[RT.UseTimeSlice])}']]," for z_ed in (z_zef_function | out_rels[RT.Binding] | collect) ],
f")\n"
        ] | filter[lambda x: x is not None] | collect
        ))    
    assert isinstance(z_func, ZefRef) and ET(z_func)==ET.ZEF_Function
    fct_body = z_func | Out[RT.PythonSourceCode] | value | collect
    assert isinstance(fct_body, str)       
    return  _make_header(z_func) + fct_body + "\n"*8






# Monkey patching ZefRef to support the call syntax
def _overloaded_zefref_call(self, *args, **kwargs):
    """
    Function that will be monkey patched in upon zefdb init for the ZefRef type object.
    Note that executing z.__call__(z) where z is a ZefRef actually calls type(z).__call__(z)
    self will be the ZefRef"""
    
    if rae_type(self) != ET.ZEF_Function: 
        raise TypeError(f'The call operator for a ZefRef can only be used if for zef functions. It was called from {self}')
    try:    # it's cheaper to ask for forgiveness
        fct = _local_compiled_zef_functions[time_resolved_hashable(self)]
    except KeyError:
        # if its not in the function cache dict, compile it
        fct = compile_zef_function(self)
    try:
        return_val = fct(*args, **kwargs)
    except Exception as exc:
        print(f"Error in executing zef function: {exc}. Reraising...")
        raise exc
    return return_val

from ..pyzef import main 
main.ZefRef.__call__ = _overloaded_zefref_call

def abstract_entity_call(entity, *args, **kwargs):
    from .abstract_raes import Entity
    if not isinstance(entity, Entity): 
        raise TypeError(f'Trying to call using abstract entity but {entity} was given instead')
    try:    # it's cheaper to ask for forgiveness
        fct = _local_compiled_zef_functions[entity.d['uid']]
    except KeyError:
        raise Exception("The abstract entity didn't have any cached compiled function.")
        # TODO: Here gather the ZefFunction and compile it!
        # fct = compile_zef_function(self)
    try:
        return_val = fct(*args, **kwargs)
    except Exception as exc:
        print(f"Error in executing zef function: {exc}. Reraising...")
        raise exc
    return return_val

# Monkey patching ZefRef to support the currying syntax
# def _overloaded_zefref_getitem(self, arg):
#     """
#     Function that will be monkey patched in upon zefdb init for the ZefRef type object.
#     self will be the ZefRef"""
    
#     if rae_type(self) != ET.ZEF_Function: 
#         raise TypeError(f'The [] syntax for a ZefRef can only be used if for zef functions. It was called from {self}')

#     from ._ops import call
#     return call[self][arg]

# from ..pyzef import main 
# main.ZefRef.__getitem__ = _overloaded_zefref_getitem



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os, sys
from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.sdist import sdist

import configparser

import subprocess

# This setup has been stolen and merged from a few places... it really needs to
# be fixed up.
class CMakeExtension(Extension):
    def __init__(self, name, cmake_lists_dir='.', cmake_path=[], cmake_args=[], **kwa):
        Extension.__init__(self, name, sources=[], **kwa)
        self.cmake_lists_dir = os.path.abspath(cmake_lists_dir)
        self.cmake_path = cmake_path
        self.cmake_args = cmake_args

class cmake_build_ext(build_ext):
    def build_extensions(self):
        # Ensure that CMake is present and working
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError('Cannot find CMake executable')

        for ext in self.extensions:

            extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
            cfg = 'Release'

            this_build_dir = os.path.join(self.build_temp, ext.name)
            # this_install_dir = os.path.join(self.build_temp, ext.name + "_install")

            os.makedirs(this_build_dir, exist_ok=True)
            # os.makedirs(this_install_dir, exist_ok=True)

            cmake_path_s = ':'.join(os.path.abspath(x) for x in ext.cmake_path)
            # Use ninja if available
            ret = subprocess.call(['which', 'ninja'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            if ret == 0:
                cmake_env = {"CMAKE_GENERATOR": "Ninja"}
                # The update after the initial value will allow an external caller to override this option if they want.
                cmake_env.update(os.environ)
            else:
                cmake_env = None
            cmake_args = [
                f'-DCMAKE_BUILD_TYPE={cfg}',
                # Hint CMake to use the same Python executable that
                # is launching the build, prevents possible mismatching if
                # multiple versions of Python are installed
                '-DPython3_EXECUTABLE={}'.format(sys.executable),
                # Add other project-specific CMake arguments if needed
                # ...
                f"-DCMAKE_PREFIX_PATH={cmake_path_s}",
                # f"-DPython_ROOT_DIR={sys.base_prefix}",

                f"-DCMAKE_INSTALL_PREFIX={extdir}",
                "-DCMAKE_OSX_DEPLOYMENT_TARGET=10.15",
            ] + ext.cmake_args

            # For debugging CI builds
            # subprocess.call(['set'], shell=True, env=cmake_env)

            # Config
            subprocess.check_call(['cmake', ext.cmake_lists_dir] + cmake_args,
                                  cwd=this_build_dir, env=cmake_env)

            # Build
            subprocess.check_call(['cmake', '--build', '.', '--config', cfg, '-v'],
                                  cwd=this_build_dir, env=cmake_env)

            # Install to the lib dir
            subprocess.check_call(['cmake', '--install', '.'],
                                  cwd=this_build_dir, env=cmake_env)
            
            


class override_sdist(sdist):
    def make_release_tree(self, base_dir, files):
        import shutil

        sdist.make_release_tree(self, base_dir, files)

        print("* Adding license to sdist")
        shutil.copy("../LICENSE", os.path.join(base_dir, "LICENSE"))

        print("* Adding libzef to sdist")

        # Adding in the libzef core too, using git to obtain the list of files.
        target_dir = os.path.join(base_dir, "libzef")
        print("Target dir is", target_dir)
        # if os.path.exists(target
        os.makedirs(target_dir)
        print("** Copying all of the files found with git ls-files across using tar packing/unpacking")
        filelist = subprocess.Popen(["git", "ls-files", "."], cwd="../core", stdout=subprocess.PIPE)
        tarcreate = subprocess.Popen(["tar", "-cvf", "-", "-T", "-"], cwd="../core", stdin=filelist.stdout, stdout=subprocess.PIPE)
        subprocess.check_call(["tar", "-xvf", "-"], cwd=target_dir, stdin=tarcreate.stdout)

        print("** Adding in the get_zeftypes.py script manually")
        # We need to include the get_zeftypes.py file along with the distribution
        shutil.copy("../scripts/get_zeftypes.py", os.path.join(target_dir, "scripts"))

        print("** Adding in the LICENSE file")
        shutil.copy("../LICENSE", target_dir)

        print("** Updating the setup.cfg file")
        # We now overwrite the setup.cfg file to point at this new directory
        setup_cfg = os.path.join(base_dir, "setup.cfg")
        parser = configparser.ConfigParser()
        with open(setup_cfg) as cfg_file:
            parser.read_file(cfg_file)
        parser["libzef"]["location"] = "libzef"
        parser["libzef"]["kind"] = "source"
        with open(setup_cfg, "w") as cfg_file:
            parser.write(cfg_file)





# This here is to create the declaration of the extension used in th build.
# Unfortunately it is run, and is very spammy, even if this file is used for
# other commands like sdist.

root = os.getcwd()
setup_cfg = os.path.join(root, "setup.cfg")
parser = configparser.ConfigParser()
with open(setup_cfg, "r") as cfg_file:
    parser.read_file(cfg_file)
libzef_location = parser.get("libzef", "location", fallback=None)
if libzef_location is None:
    libzef_location = os.environ.get("LIBZEF_LOCATION", None)
libzef_kind = parser.get("libzef", "kind", fallback="guess")
if libzef_kind == "guess":
    if libzef_location is None:
        # We first try without the default location to see if anything is found
        ret = subprocess.call(['cmake', '-P', 'lookforzef.cmake'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        if ret == 0:
            libzef_kind = "binary"
        else:
            libzef_location = "../core"

    if libzef_kind == "guess":
        ret = subprocess.call(['cmake', '-DCMAKE_PREFIX_PATH=' + libzef_location, '-P', 'lookforzef.cmake'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        if ret == 0:
            libzef_kind = "binary"
        elif os.path.exists(os.path.join(libzef_location, "zefDBConfig.cmake.in")):
            libzef_kind = "source"
        else:
            raise Exception(f"Can't guess what kind of libzef exists at '{libzef_location}' looking from the cwd of '{os.getcwd()}'. If this is a relative directory, and you are installing via `pip wheel ...` or similar, then you should specify it in either the setup.cfg, or via the `LIBZEF_LOCATION` environment variable, as a absolute directory.")

    print(f"libzef location was guessed to be '{libzef_kind}' from using directory '{libzef_location}' (although it may be on the cmake path instead)")
    
if libzef_location is None:
    libzef_paths = []
else:
    libzef_paths = [libzef_location]

if libzef_kind == "binary":
    print(f"Using libzef as a binary with locations: {libzef_paths}")
    pyzef_ext = CMakeExtension("zef.pyzef", "pyzef",
                    cmake_path=libzef_paths,)
elif libzef_kind == "source":
    print(f"Bundling libzef from source in locations: {libzef_paths}")
    pyzef_ext = CMakeExtension("zef.pyzef", "pyzef",
                    cmake_path=libzef_paths,
                    cmake_args=["-DLIBZEF_PYZEF_BUNDLED=TRUE"])
else:
    raise Exception(f"Don't understand libzef_kind of '{libzef_kind}'")




# I HATE PYTHON PACKAGING
sys.path += [""]
import versioneer

with open("requirements.txt") as file:
    reqs = [z.strip() for z in file.readlines() if z.strip()]


setup(
    name="zef",
    author="ZefHub.io",
    python_requires=">=3.8",
    version=versioneer.get_version(),
    packages=find_packages(),
    install_requires=reqs,
    ext_modules=[pyzef_ext],
    cmdclass = versioneer.get_cmdclass({
        'build_ext': cmake_build_ext,
        'sdist': override_sdist,
    })
)





# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import structlog

# Trying to keep a config that doesn't interfer with any user's global settings
# for structlog if they happen to use it.
log = structlog.wrap_logger(None,
    processors=[
        # structlog.stdlib.filter_by_level,
        # structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        # structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.TimeStamper(fmt="iso"),
        # The following only does anything if exc_info is passed as a keyword
        structlog.processors.format_exc_info,
        structlog.processors.UnicodeDecoder(),
        structlog.dev.ConsoleRenderer(colors=True),
    ]
)



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zefdb import *
from zefdb.zefops import *
from zefdb import zefui

from rx.subject import ReplaySubject, Subject

# TODO: This will go into the graph crawler I'm guessing...
component_lookups = {}

def make_component(z_root, sid):
    z_template = component_lookups[ET(z_root)]

    # This is weird naming for the purpose here, will likely change in the future.
    zefui.start_backend_session(z_root, z_template, sid)


#######


def generate_template_graph_programmatically():
    g = Graph()
    with Transaction(g):
        I = lambda *args: instantiate(*args, g)

        input_component = I(ET.Input) | attach[[
            (RT.Value, "RAW(empty input)"),
            (RT.ReadOnly, False),
            (RT.Placeholder, "RAW(Enter text)")
        ]]

        update_component = I(ET.SendMessage) | attach[RT.MessageTemplate, """
output = {"components":{z_component_root|uid: {"component": "input"}}}
"""]

        update_value = I(ET.SendMessage) | attach[RT.MessageTemplate, """
maybe_ae = z_component_root >> O[RT.Value]
if maybe_ae is None:
    # This will be replaced by using the default value in the template graph.
    output = {"components":{z_component_root|uid: {"value": f"RAW(empty input)"}}}
else:
    output = {"components":{z_component_root|uid: {"value": f"STATE({maybe_ae|uid})"}}}
"""]

        update_onchange = I(ET.SendMessage) | attach[RT.MessageTemplate, """
value_present = z_component_root | has_out[RT.Value]
readonly = z_component_root >> O[RT.ReadOnly]
readonly = None if readonly is None else readonly|value
readonly = False if readonly is None else readonly

if value_present and not readonly:
    output = {"components":{z_component_root|uid: {
        "onChange": f"OPTIMISTIC({z_component_root>>RT.Value|uid})",
        "class": "",
}}}
else:
    output = {"components":{z_component_root|uid: {
        "onChange": None,
        "class": "readonly"
}}}
"""]

        update_placeholder  = I(ET.SendMessage) | attach[RT.MessageTemplate,  """
maybe_ae = z_component_root >> O[RT.Placeholder]
if maybe_ae is None:
    output = {"components":{z_component_root|uid: {"placeholder": f"RAW(Enter text)"}}}
else:
    output = {"components":{z_component_root|uid: {"placeholder": f"STATE({maybe_ae|uid})"}}}
"""]

        input_component | attach[[
            (RT.ZefUI_OnInstantiation, update_component),
        ]]

        (input_component > RT.Value) | attach[[
            (RT.ZefUI_OnInstantiation, update_value),
            (RT.ZefUI_OnInstantiation, update_onchange),
            (RT.ZefUI_OnTermination, update_value),
            (RT.ZefUI_OnTermination, update_onchange),
        ]]

        (input_component > RT.ReadOnly) | attach[[
            (RT.ZefUI_OnInstantiation, update_onchange),
            (RT.ZefUI_OnTermination, update_onchange),
        ]]
        (input_component >> RT.ReadOnly) | attach[[
            (RT.ZefUI_OnValueAssignment, update_onchange),
        ]]

        (input_component > RT.Placeholder) | attach[[
            (RT.ZefUI_OnInstantiation, update_placeholder),
            (RT.ZefUI_OnTermination, update_placeholder),
        ]]

    component_lookups[ET.Input] = input_component|now
    global stored_template_g
    stored_template_g = g




# To be redone in zefui

def create_zefui_streams(session_id=None):    
    # if no session_id is provided: check if there is only one session
    if session_id is None:        
        d = zefui._state['sessions']
        if len(d) != 1:
            raise RuntimeError(f"hook_up_streams_listeners called with no session id specified, but there are {len(d)} zefui sessions. It is not clear which one to choose.")
        session_id = next(iter(d.keys()))        
    
    ui_model = zefui._state['sessions'][session_id]['ui_model']
        
    # TODO: Instead of a replay subject, need to have a way to get the full current state.
    outgoing_stream = ReplaySubject()
    push_batched_msgs_from_tx_to_stream = zefui.make_push_batched_msgs_from_tx_to_stream(outgoing_stream)
    # this function is triggered after all individual subscriptions fire, once upon each tx closing on the ui_model
    ui_model | subscribe[push_batched_msgs_from_tx_to_stream]      
    # After realizing this too late: we also need to trigger this fct once here, since no tx is closing when setting all up
    # the arg passed to a general graph subscription is the root node in the respective time slice
    push_batched_msgs_from_tx_to_stream(ui_model[42]|now)
    # -----------------              hook up part to act on the messages coming in from the FE through the WS             ------------------
    # execute the zefui user-registered fcts or transform the messages and push them into the stream
    
    incoming_stream = Subject()
    
    def handle_msg_from_fe(msg: dict):
        """check what kind of message this is: look in the local dict whether 
        there are callbacks / cb streams registered for this kind of msg."""
        print(f"@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ handle_msg_from_fe: new msg received from FE: {msg}")
        # TODO!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    incoming_stream.pipe(
        #op.map(lambda msg_str: json.loads(msg_str))        
        ).subscribe(handle_msg_from_fe)

    return outgoing_stream, incoming_stream







import rx.operators as rxop
from rx.scheduler.eventloop import AsyncIOScheduler
from rx.core import notification as rxnotify
import asyncio

"""Mimic subscribe of an rx stream, but where the subscription function is an
asyncio coroutine AND these coroutines are guaranteed to be executed in order of
the stream.

Note: if order of execution is not required, then something like:
stream.pipe(rxop.map(lambda x: rxop.from_future(asyncio.create_task(func(x)))),
            rxop.merge_all()
           ).subscribe()
could be used instead.

Or
stream.subscribe(lambda x: asyncio.create_task(func(x)))
"""
async def sync_map_rx_stream(func, stream):
    queue = asyncio.Queue()

    disposable = stream.pipe(
        rxop.materialize()
    ).subscribe(
        on_next=lambda i: queue.put_nowait(i),
        scheduler=AsyncIOScheduler(loop=asyncio.get_event_loop())
    )

    while True:
        i = await queue.get()
        if isinstance(i, rxnotify.OnNext):
            await func(i.value)
            queue.task_done()
        else:
            disposable.dispose()
            if isinstance(i, rxnotify.OnError):
                raise(Exception(i.value))
            break



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *
import zef


class MyTestCase(unittest.TestCase):

    def test_basics(self):
        g = Graph()

        r = [
            ET.Person["joe"],
            (Z["joe"], RT.FirstName, "Joe"),
            (Z["joe"], RT.LastName, "Bloggs"),
            (Z["joe"], RT.NickName, "Guy"),
            (Z["joe"], RT.NickName, "Joey"),

            ET.Pet["spot"],
            (Z["spot"], RT.Name, "Spot"),
            (Z["joe"], RT.HasPet["joe-spot"], Z["spot"]),

            (Z["joe-spot"], RT.Date, Time("12:34 May 2020")),
        ] | transact[g] | run

        self.assertEqual(r["joe"], g | now | all[ET.Person] | only | collect)
        self.assertEqual(r["spot"], g | now | all[ET.Pet] | only | collect)

        self.assertEqual(r["joe"] | Out[RT.FirstName] | value | collect, "Joe")
        self.assertEqual(r["joe"] | Out[RT.HasPet] | collect, r["spot"])
        self.assertEqual(r["joe"] | Outs[RT.NickName] | length | collect, 2)

        z_joe = r["joe"]
        r2 = [
            (r["joe"], RT.NickName, "Jay"),
            r["joe"] | Out[RT.LastName] | assign_value["Smith"],
        ] | transact[g] | run

        z2_joe = now(z_joe)

        self.assertEqual(z2_joe | Out[RT.LastName] | value | collect, "Smith")
        self.assertEqual(z2_joe | Outs[RT.NickName] | length | collect, 3)

        r3 = [
            *[x | terminate for x in z2_joe | Outs[RT.NickName] | collect],
        ] | transact[g] | run

        z3_joe = now(z2_joe)

        self.assertEqual(z3_joe | Outs[RT.NickName] | length | collect, 0)
        self.assertFalse(z3_joe | has_out[RT.NickName] | collect)
        

    def test_merges(self):
        g = Graph()

        r = [
            ET.Person["joe"],
            (Z["joe"], RT.FirstName, "Joe"),
            (Z["joe"], RT.LastName, "Bloggs"),
        ] | transact[g] | run

        z_joe = r["joe"]

        g2 = Graph()

        r2 = [
            z_joe,
            (z_joe, RT.Owns, ET.Account["acc"]),
            (Z["acc"], RT.Balance, QuantityFloat(100, EN.Unit.dollars)),
        ] | transact[g2] | run

        self.assertTrue(uid(z_joe|to_ezefref) in g2)
        self.assertEqual(length(g | now | all[ET.Person]), 1)

        # TODO: Replace with proper lookup later
        z2_joe = g2[uid(z_joe|to_ezefref)] | In[BT.ORIGIN_RAE_EDGE] | target | now | collect
        r3 = [
            (z2_joe, RT.FromMerge, True)
        ] | transact[g] | run

        self.assertEqual(length(g | now | all[ET.Person]), 1)
        self.assertEqual(z_joe | now | Out[RT.FromMerge] | value | collect, True)

        z3_joe = z_joe | g2 | run

        self.assertEqual(to_ezefref(z2_joe), to_ezefref(z3_joe))

    def test_multiple_relation_notation(self):
        g = Graph()

        r = [
            ([ET.Person["joe"], ET.Person["john"]], RT.Owns, [ET.Pet["cat"], ET.Pet["dog"]]),
        ] | transact[g] | run

        self.assertEqual(g | now | all[RT.Owns] | length | collect, 4)
        self.assertTrue(has_relation(r["joe"], RT.Owns, r["cat"]))
        self.assertTrue(has_relation(r["joe"], RT.Owns, r["dog"]))
        self.assertTrue(has_relation(r["john"], RT.Owns, r["cat"]))
        self.assertTrue(has_relation(r["john"], RT.Owns, r["dog"]))

        g = Graph()
        r = [
            (ET.Person, RT.Owns, [ET.Pet["a"], ET.Pet["b"]]),
        ] | transact[g] | run

        self.assertEqual(g | now | all[RT.Owns] | length | collect, 2)
        self.assertTrue(has_relation(g | now | all[ET.Person] | single | collect, RT.Owns, r["a"]))
        self.assertTrue(has_relation(g | now | all[ET.Person] | single | collect, RT.Owns, r["b"]))

        g = Graph()
        r = [
            (ET.Person["joe"], [(RT.FirstName, "Joe"),
                                (RT.LastName, "Bloggs")])
        ] | transact[g] | run

        self.assertEqual(r["joe"] | Out[RT.FirstName] | value | collect, "Joe")
        self.assertEqual(r["joe"] | Out[RT.LastName] | value | collect, "Bloggs")

    def test_relation_relations(self):
        g = Graph()
        z = ET.Example | g | run
        x, y, z = (z, RT.Something, 3) | g | run
        x, y, z = (y, RT.Something, 3) | g | run
        g | now | all[ET.Example] | last | out_rel[RT.Something] | Out[RT.Something] | collect

    def test_tagging(self):
        g = Graph()

        z = ET.Person | g | run
        r = [
            ET.Person["joe"],
            Z["joe"] | tag["secret"],
            z | tag["first"]
        ] | transact[g] | run

        self.assertEqual(r["joe"], g | now | get["secret"] | collect)
        self.assertEqual(z | now | collect, g | now | get["first"] | collect)

if __name__ == '__main__':
    unittest.main()
